#__________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Nov 2018
#
# _________________________________________________________________________
# various tools machine learning using Keras platform
#

from datetime import datetime, timedelta
import numpy as np
import tensorflow as tf
from tensorflow.keras.utils import Sequence, to_categorical

import dgbpy.keystr as dgbkeys
from dgbpy import hdf5 as dgbhdf5
from dgbpy import dgbkeras
from dgbpy import keras_fix

class TrainingSequence(Sequence):
  def __init__(self,trainbatch,forvalidation,model,exfilenm=None,batch_size=1,\
               with_augmentation=True,tempnm=None):
      self._trainbatch = trainbatch
      self._forvalid = forvalidation
      self._model = model
      self._nrdone = -1
      self._tempnm = tempnm
      self._lastsaved = datetime.now()
      self._batch_size = batch_size
      self._augmentation = with_augmentation
      self._channels_format = dgbkeras.get_data_format(model)
      self._infos = self._trainbatch[dgbkeys.infodictstr]
      self._data_IDs = []
      if exfilenm == None:
        self._exfilenm = self._infos[dgbkeys.filedictstr]
      else:
        self._exfilenm = exfilenm
      self._nrclasses = 0
      if self._infos[dgbkeys.classdictstr]:
        self._nrclasses = dgbhdf5.getNrClasses( self._infos )
        if dgbhdf5.isImg2Img(self._infos) and self._nrclasses <= 2:
          self._nrclasses = 0

  def __len__(self):
      return int(np.floor(len(self._data_IDs)/float(self._batch_size)))

  def set_chunk(self,ichunk):
      infos = self._infos
      nbchunks = len(infos[dgbkeys.trainseldicstr])
      if nbchunks > 1:
          from dgbpy import mlapply as dgbmlapply
          trainbatch = dgbmlapply.getScaledTrainingDataByInfo( infos,
                                                 flatten=False,
                                                 scale=True, ichunk=ichunk )
      else:
          trainbatch = self._trainbatch
          
      if self._forvalid:
          if not dgbkeys.xvaliddictstr in trainbatch or \
             not dgbkeys.yvaliddictstr in trainbatch:
              return False
          x_data = trainbatch[dgbkeys.xvaliddictstr]
          y_data = trainbatch[dgbkeys.yvaliddictstr]
      else:
          if not dgbkeys.xtraindictstr in trainbatch or \
             not dgbkeys.ytraindictstr in trainbatch:
              return False
          x_data = trainbatch[dgbkeys.xtraindictstr]
          y_data = trainbatch[dgbkeys.ytraindictstr]
      model = self._model
      self._x_data = dgbkeras.adaptToModel( model, x_data )
      if len(y_data.shape) > 2:
          self._y_data = dgbkeras.adaptToModel( model, y_data )
      else:
          self._y_data = y_data
      inp_shape = self._x_data.shape[1:]
      if self._augmentation and len(inp_shape) == 4:
          if self._channels_format == 'channels_first':
              self._rotdims = (2,3)
              cubesz = self._x_data.shape[2:4]
          else:
              self._rotdims = (1,2)
              cubesz = self._x_data.shape[1:3]
          if cubesz[0] == cubesz[1]:
              self._rot = range(4)
              self._rotidx = self._rot
          else:
              self._rot = range(2)
              self._rotidx = range(0,4,2)
      elif self._augmentation and len(inp_shape) == 3:
          self._rot = range(2)
          self._rotidx = range(0,4,2)
          if self._channels_format == 'channels_first':
              self._rotdims = 2
          else:
              self._rotdims = 1
      else:
          self._rot = range(1)
      self._data_IDs = range(len(self._x_data)*len(self._rot))
      self.on_epoch_end()
      return True

  def on_epoch_end(self):
      self._nrdone = self._nrdone+1
      if self._tempnm != None and self._nrdone > 0:
          now = datetime.now()
          if now - self._lastsaved > timedelta(minutes=10):
              dgbkeras.save( self._model, self._tempnm )
              self._lastsaved = now
      self._indexes = np.arange(len(self._data_IDs))
      np.random.shuffle(self._indexes)
      
  def __getitem__(self, index):
      islast = index==(len(self)-1)
      bsize = self._batch_size
      if islast:
        indexes = self._indexes[index*bsize:]
      else:
        indexes = self._indexes[index*bsize:(index+1)*bsize]
      data_IDs_temp = [self._data_IDs[k] for k in indexes]
      return self.__data_generation(data_IDs_temp)
    
  def __data_generation(self, data_IDs_temp):
      x_data = self._x_data
      y_data = self._y_data
      inp_shape = x_data.shape[1:]
      out_shape = y_data.shape[1:]
      nrpts = len(data_IDs_temp)
      X = np.empty( (nrpts,*inp_shape), dtype=x_data.dtype )
      Y = np.empty( (nrpts,*out_shape), dtype=y_data.dtype )
      nrrot = len(self._rot)
      if nrrot == 1:
          X = x_data[data_IDs_temp]
          Y = y_data[data_IDs_temp]
      else:
          iindex,frem = np.divmod(data_IDs_temp,nrrot)
          n = 0
          rotdims = self._rotdims
          flip2d = not isinstance( rotdims, tuple )
          for j,k in zip(self._rot,self._rotidx):
              rotidx = iindex[frem==j]
              m = len(rotidx)
              if m > 0:
                  wrrg = range(n,n+m)
                  if flip2d:
                      if k == 0:
                          X[wrrg] = x_data[rotidx]
                      else:
                          X[wrrg] = np.fliplr(x_data[rotidx])
                  else:
                      X[wrrg] = np.rot90(x_data[rotidx],k,rotdims)
                  if len(y_data.shape) > 2:
                      if flip2d:
                          if k == 0:
                              Y[wrrg] = y_data[rotidx]
                          else:
                              Y[wrrg] = np.fliplr(y_data[rotidx])
                      else:
                          Y[wrrg] = np.rot90(y_data[rotidx],k,rotdims)
                  else:
                      Y[wrrg] = y_data[rotidx]
                  n = wrrg.stop
      if self._nrclasses > 0:
          Y = to_categorical(Y,self._nrclasses)
      return (X, Y, [None])

import importlib
import pkgutil
import inspect

from abc import ABC, abstractmethod
from keras import backend
from pathlib import Path
from enum import Enum

class DataPredType(Enum):
  Continuous = 'Continuous Data'
  Classification = 'Classification Data'
  Segmentation = 'Segmentation'
  Any = 'Any'

class OutputType(Enum):
  Pixel = 1
  Image = 2
  Any = 3
    
class DimType(Enum):
  D1 = 1
  D2 = 2
  D3 = 3
  Any = 4

class UserModel(ABC):
  """Abstract base class for user defined Keras machine learning models
  
  This module provides support for users to add their own machine learning
  models to OpendTect.

  It defines an abstract base class. Users derive there own model classes from this base
  class and implement the _make_model static method to define the structure of the keras model.
  The users model definition should be saved in a file name with "mlmodel_" as a prefix and be 
  at the top level of the module search path so it can be discovered.
  
  The "mlmodel_" class should also define some class variables describing the class:
  uiname : str - this is the name that will appear in the user interface
  uidescription : str - this is a short description which may be displayed to help the user
  predtype : DataPredType enum - type of prediction (must be member of DataPredType enum)
  outtype: OutputType enum - output shape type (OutputType.Pixel or OutputType.Image)
  dimtype : DimType enum - the input dimensions supported by model (must be member of DimType enum)

  Examples
  --------
    from dgbpy.keras_classes import UserModel, DataPredType, OutputType, DimType
  
    class myModel(UserModel):
      uiname = 'mymodel'
      uidescription = 'short description of model'
      predtype = DataPredType.Classification
      outtype = OutputType.Pixel
      dimtype = DimType.D3
      
      def _make_model(self, input_shape, nroutputs, learnrate, data_format):
        inputs = Input(input_shape)
        conv1 = Conv3D(2, (3,3,3), activation='relu', padding='same')(inputs)
        conv1 = Conv3D(2, (3,3,3), activation='relu', padding='same')(conv1)
        pool1 = MaxPooling3D(pool,size=(2,2,2))(conv1)
        ...
        conv8 = Conv3D(1, (1,1,1,), activation='sigmoid')(conv7)
      
        model = Model(inputs=[inputs], outputs=[conv8])
        model.compile(optimizer = Adam(lr = 1e-4), loss = cross_entropy_balanced, metrics = ['accuracy'])
        return model
      
    
  """
  mlmodels = []
  
  def __init__(self, ):
    self._learnrate = None
    self._nroutputs = None
    self._data_format = None
    self._model = None

  @staticmethod
  def findModels():
    """Static method that searches the PYTHONPATH for modules containing user
    defined Keras machine learning models (UserModels).
    
    The module name must be prefixed by "mlmodel_". All subclasses of the
    UserModel base class is each found module will be added to the mlmodels
    class variable.
    """

    mlm = []

    for _, name, ispkg in pkgutil.iter_modules(path=[Path(__file__).parent.absolute()]):
      if name.startswith("mlmodel_"):
        module = importlib.import_module('.'.join(['dgbpy',name]))
        clsmembers = inspect.getmembers(module, inspect.isclass)
        for (_, c) in clsmembers:
          if issubclass(c, UserModel) & (c is not UserModel):
            mlm.append(c())
        
    for _, name, ispkg in pkgutil.iter_modules():
      if name.startswith('mlmodel_'):
        module = importlib.import_module(name)
        clsmembers = inspect.getmembers(module, inspect.isclass)
        for (_, c) in clsmembers:
          if issubclass(c, UserModel) & (c is not UserModel):
            mlm.append(c())
    return mlm
  
  @staticmethod
  def findName(modname):
    """Static method that searches the found UserModel's for a match with the
    uiname class variable
    
    Parameters
    ----------
    modname : str
    Name (i.e. uiname) of the UserModel to search for.
    
    Returns
    -------
    an instance of the class with the first matching name in the mlmodels
    list or None if no match is found
    
    """
    return next((model for model in UserModel.mlmodels if model.uiname == modname), None)
  
  @staticmethod
  def getModelsByType(pred_type, out_type, dim_type):
    """Static method that returns a list of the UserModels filtered by the given
    prediction, output and dimension types
    
    Parameters
    ----------
    pred_type: DataPredType enum
    The prediction type of the model to filter by
    out_type: OutputType enum
    The output shape type of the model to filter by
    dim_type: DimType enum
    The dimensions that the model must support
    
    Returns
    -------
    a list of matching model or None if no match found
    
    """
    if isinstance(pred_type, DataPredType) and isinstance(out_type, OutputType) and\
       isinstance(dim_type, DimType) :
      return [model for model in UserModel.mlmodels \
          if (model.predtype == pred_type or pred_type == DataPredType.Any) and\
	     (model.outtype == out_type or out_type == OutputType.Any) and\
             (model.dimtype == dim_type or model.dimtype == DimType.Any)]
    return None

  @staticmethod
  def getNamesByType(pred_type, out_type, dim_type):
      models = UserModel.getModelsByType(pred_type, out_type, dim_type)
      return [model.uiname for model in models]
  
  @staticmethod
  def isPredType( modelnm, pred_type ):
      models = UserModel.getModelsByType( pred_type, OutputType.Any, DimType.Any )
      for mod in models:
          if mod.uiname == modelnm:
              return True
      return False
  
  @staticmethod
  def isOutType( modelnm, out_type ):
      models = UserModel.getModelsByType( DataPredType.Any, out_type, DimType.Any )
      for mod in models:
          if mod.uiname == modelnm:
              return True
      return False
  
  @staticmethod
  def isClassifier( modelnm ):
      return UserModel.isPredType( modelnm, DataPredType.Classification )
  
  @staticmethod
  def isRegressor( modelnm ):
      return UserModel.isPredType( modelnm, DataPredType.Continuous )  
  
  @staticmethod
  def isImg2Img( modelnm ):
      return UserModel.isOutType( modelnm, OutputType.Image )
  
  @abstractmethod
  def _make_model(self, input_shape, nroutputs, learnrate, data_format):
    """Abstract static method that defines a machine learning model.
    
    Must be implemented in the user's derived class
    
    Parameters
    ----------
    input_shape : tuple
    Defines input data shape in the Keras default data_format for the current backend.
    For the TensorFlow backend the default data_format is 'channels_last'
    nroutputs : int (number of discrete classes for a classification)
    Number of outputs
    learnrate : float
    The step size applied at each iteration to move toward a minimum of the loss function
    
    Returns
    -------
    a compiled keras model
    
    """
    pass

  def model(self, input_shape, nroutputs, learnrate, data_format='channels_first'):
    """Creates/returns a compiled keras model instance
    
    Parameters
    ----------
    input_shape : tuple
    Defines input data shape arranged as per the data_format setting. 
    nroutputs : int (number of discrete classes for a classification)
    Number of outputs
    learnrate : float
    The step size applied at each iteration to move toward a minimum of the loss function
    data_format: str
    The data format used. The machine learning plugin uses 'channels_first' data_format.
    
    Returns
    -------
    a compiled keras model
    
    """
    modshape = input_shape
    if data_format=='channels_first' and tf.keras.backend.image_data_format()=='channels_last':
      modshape = (*input_shape[1:], input_shape[0])
    elif data_format=='channels_last' and tf.keras.backend.image_data_format()=='channels_first':
      modshape = (input_shape[-1], *input_shape[0:-1])
      
    newmodel = self._model is None or modshape != self._model.input_shape or \
                nroutputs != self._nroutputs or learnrate != self._learnrate
    if  newmodel:
      self._nroutputs = nroutputs
      self._learnrate = learnrate
      self._model = self._make_model(modshape,nroutputs,learnrate)
      self._data_format = dgbkeras.get_data_format( self._model )
    return self._model
  
UserModel.mlmodels = UserModel.findModels()



"""Tools for managing OS commands

Copyright (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
  * AUTHOR : A. Huck
  * DATE   : Apr 2019

"""

import sys
import os
import psutil
import signal
import subprocess
import json
import threading

from odpy.common import isWin, getExecPlfDir, get_log_stream, get_std_stream, std_msg, sTimeUnitString, log_msg

def getODCommand(execnm,args=None):
  """OpendTect command

  For a given OpendTect executable, get its full command line

  Parameters:
    * execnm (list): [Executable name]
    * arg (dict, optional):
      Dictionary with the members 'dtectdata' and 'survey' as 
      single element lists, and/or 'dtectexec' (see odpy.getODSoftwareDir)

  Returns:
    * list: OpendTect command line

  Notes:
    The two standard arguments providing the current survey location
    will be added to the command line arguments.

  Examples:
    >>> args = {
      'dtectdata': ['D:\\ODData'],
      'survey': ['F3_Demo']
    }

    >>> getODCommand()
    ['C:\\Program Files\\OpendTect\\6.6.0\\bin\win64\Release\\od_process_attrib',
    '--dataroot',
    'D:\\ODData',
    '--survey',
    'F3_Demo']

  """

  cmd = list()
  cmd.append( os.path.join(getExecPlfDir(args),execnm) )
  return appendDtectArgs( cmd, args )

def appendDtectArgs( cmd, args=None ):
  """Append OpendTect arguments

  Parameters:
    * cmd (list): List to which the returned elements will be added
    * arg (dict, optional):
      Dictionary with the members 'dtectdata' and 'survey' as 
      single element lists, and/or 'dtectexec' (see odpy.getODSoftwareDir)

  Returns:
    * list: Input list with the added data root and survey directory name if provided

  Examples:
    >>> args = {
      'dtectdata': ['D:\\ODData'],
      'survey': ['F3_Demo']
    }

    >>> appendDtectArgs( args )
    ['--dataroot', 'D:\\ODData', '--survey', 'F3_Demo']

  """

  if args == None:
    return cmd
  if 'dtectdata' in args:
    cmd.append( '--dataroot' )
    cmd.append( args['dtectdata'][0] )
  if 'survey' in args:
    cmd.append( '--survey' )
    cmd.append( args['survey'][0] )
  return cmd

def getPythonExecNm():
  """Python executable name

  shortcut to sys.executable
  
  """

  return sys.executable

def getPythonCommand(scriptfile,posargs=None,dict=None,args=None):
  """Python command line

  Parameters:
    * scriptfile (str): Python script or executable
    * posargs (list, optional): 
      Positional arguments for that script (default is None)
    * dict (str, optional): Value for the argument with key "--dict":
      String from which a json object can be parsed (default is None)
    * args (dict):
      Dictionary with the members 'dtectdata' and 'survey' as 
      single element lists, and/or 'dtectexec' (see odpy.getODSoftwareDir).
      The dictionary may also contain the members 'proclog' and 'syslog' as 
      strings, each being the full path to a log file.

  Returns:
    * list: 
      Fully formed command line to be executed by the python executable

  """

  cmd = list()
  cmd.append( getPythonExecNm() )
  cmd.append( scriptfile )
  cmd = appendDtectArgs( cmd, args )
  cmd.append( '--dtectexec' )
  cmd.append( getExecPlfDir(args) )
  if args != None and 'proclog' in args:
    cmd.append( '--proclog' )
    cmd.append( args['proclog'] )
  if args != None and 'syslog' in args:
    cmd.append( '--syslog' )
    cmd.append( args['syslog'] )
  for posarg in posargs:
    cmd.append( posarg )
  if dict != None:
    cmd.append( '--dict' )
    cmd.append( json.dumps(dict) )
  return cmd

def execCommand( cmd, background=False ):
  """Command execution

  Launches the execution of a command with its arguments.

  Parameters:
    * cmd (list): command to be executed with its arguments.
    * background (bool, optional):
      The command is forked if True (default is False).

  Returns:
    * str: 
      The stdout print of the subprocess.CompletedProcess (default)
      or a psutil.Popen object (background=True)

  Notes:
    This is an interface to the subprocess module.

  """

  if background:
    return startDetached( cmd )
  else:
    return startAndWait( cmd )

def isRunning( proc ):
  """is the process running?

  Parameters:
    * proc (psutil.Popen): Process to be checked

  Returns:
    * bool: True if the process is running.

  """

  if proc == None:
    return
  proc.poll()
  return proc.is_running()

def pauseProcess( proc ):
  """Pause a running process

  Parameters:
    * proc (psutil.Popen): Process to be paused

  Returns:
    * Set and return subprocess.Popen.returncode

  """

  if isinstance(proc,psutil.Popen):
    return proc.suspend()

def resumeProcess( proc ):
  """Resume a paused process

  Parameters:
    * proc (psutil.Popen): Process to be resumed

  Returns:
    * Set and return subprocess.Popen.returncode

  """

  if isinstance(proc,psutil.Popen):
    return proc.resume()

def printProcessTime(procnm,isstart,print_fn=std_msg,withprocline=True):
  """Print processing timestamp

  Parameters:
    * procnm (str): Process name
    * isstart (bool): Status: Start or end of this process.
    * print_fn (function, optional):
      function to be used for the printing of the timestamp
      (default is odpy.std_msg, which default to stderr)
    * withprocline (bool, optional):
      Print the timestamp preceeded by a label string (default is True)

  Examples:
    >>> printProcessTime( 'od_process', True )
    Process: 'od_process'
    Started: Mon 20 Apr 2020, 17:23:32

    >>> printProcessTime( 'od_process', False )
    Process: 'od_process'
    Finished: Mon 20 Apr 2020, 17:23:37

    >>> printProcessTime( 'od_process', False, withprocline=False )
    Finished: Mon 20 Apr 2020, 17:23:37

  """
  if withprocline:
    procstr = 'Process: \'' + procnm + '\''
    print_fn( procstr )
  if isstart:
    retstr = 'Started:'
  else:
    retstr = 'Finished:'
  print_fn( retstr, sTimeUnitString() )

def kill( proc ):
  """Terminate a process

  Parameters:
    * proc (psutil.Popen): Object view of the process

  Returns:
    * Set and return subprocess.Popen.returncode

  """

  with proc.oneshot():
    proc.resume()
    proc.terminate()
    proc.wait(3)

def startAndWait( cmd ):
  """Run a command with subprocess

  Parameters:
    * cmd (list): command to be executed with its arguments.

  Returns:
    * str:
      stdout from the executed and retrieved subprocess.CompletedProcess

  Notes:
    * The command is executed by the run function of the subprocess module.
    * stdout is captured with a PIPE from the child process.
    * stderr is captured and forwarded to odpy.syslog_logger.

  """

  try:
    completedproc = subprocess.run( cmd, check=True, stdout=subprocess.PIPE, \
                                    stderr=get_std_stream() ).stdout
  except subprocess.CalledProcessError as err:
    std_msg( 'Failed: ', err )
    raise
  return completedproc

def log_subprocess_output(pipe):
    for line in iter(pipe.readline, b''):
        log_msg(line.decode('utf-8').strip('\n'))

def startDetached( cmd ):
  """Run a command in the background with psutil

  Parameters:
    * cmd (list): command to be executed with its arguments.

  Returns:
    psutil.Popen: Object view of the process

  Notes:
    * The command is executed by the Popen function of the psutil module.
    * Since the command is run as a daemon, it does not block the execution of the script.
    * stdout is captured and forwarded to odpy.proclog_logger.
    * stderr is captured and forwarded to odpy.proclog_logger.

  """

  try:
    runningproc = None
    if isWin():
      runningproc = psutil.Popen( cmd, stdout=subprocess.PIPE, \
                                       stderr=subprocess.STDOUT )
    else:
      runningproc = psutil.Popen( cmd, start_new_session=True, \
                                       stdout=subprocess.PIPE, \
                                       stderr=subprocess.STDOUT )
    t = threading.Thread(target=log_subprocess_output, args=(runningproc.stdout,))
    t.start()
  except subprocess.CalledProcessError as err:
    std_msg( 'Failed: ', err )
    raise
  return runningproc




#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : Bert
# DATE     : August 2018
#
# IOPar tools
#

import odpy.common


def read_line( fp, bin ):
  if bin:
    words = fp.readline().decode("utf8").split(":")
  else:
    words = fp.readline().split(":")
  cleanwords = [x.strip() for x in words]
  nrwords = len(cleanwords)
  ret = list()
  if nrwords > 0:
    ret.append( cleanwords[0] )
  if nrwords > 1:
    ret.append( ':'.join(cleanwords[1:nrwords]) )
  return ret



import json

from odpy.common import isWin, log_msg
from odpy.oscommand import getODCommand, execCommand

dbmanexe = 'od_DBMan'

def getDBList(translnm,alltrlsgrps=False,exenm=dbmanexe,args=None):
  cmd = getODCommand(exenm,args)
  cmd.append( '--json' )
  if alltrlsgrps:
    cmd.append( '--all' )
  cmd.append( '--list' )
  cmd.append( translnm )
  return getDBDict( cmd )

def getInfoFromDBListByNameOrKey(nm_or_key,dblist):
  for i in range(len(dblist['Names'])):
    dbobjnm = dblist['Names'][i]
    objid = dblist['IDs'][i]
    objfmt = dblist['Formats'][i]
    if dbobjnm != nm_or_key and objid != nm_or_key:
      continue
    if 'TranslatorGroups' in dblist:
      objtrlgrp = dblist['TranslatorGroups'][i]
    if 'Types' in dblist:
      objtyp = dblist['Types'][i]
    ret = {
      'ID': objid,
      'Name': dbobjnm,
      'Format': objfmt,
    }
    if len(objtrlgrp) > 0:
      ret.update({'TranslatorGroup': objtrlgrp})
    if len(objtyp) > 0:
      ret.update({'Type': objtyp})
    return ret

def getInfoByName(objnm,translnm,exenm=dbmanexe,args=None ):
  cmd = getODCommand(exenm,args)
  cmd.append( '--json' )
  cmd.append( '--exists' )
  cmd.append( objnm )
  cmd.append( '--trl-grp' )
  cmd.append( translnm )
  ret = getDBDict( cmd )
  if not 'ID' in ret:
    return None
  return ret

def getInfoByKey(objkey,exenm=dbmanexe,args=None ):
  cmd = getODCommand(exenm,args)
  cmd.append( '--json' )
  cmd.append( '--info' )
  cmd.append( objkey )
  return getDBDict( cmd )

def getDBDict( cmd ):
  ret = execCommand( cmd )
  retstr = ret.decode('utf-8')
  if isWin():
    retstr = retstr.translate(str.maketrans({"\\": r"\\"}))
  ret = json.loads( retstr )
  if ret['Status'] != 'OK':
    log_msg( ret['Status'] )
    raise FileNotFoundError
  return ret

def getByName( dblist, retname, keystr ):
  curentryidx = dblist['Names'].index( retname )
  return dblist[keystr][curentryidx]

def getDBKeyForName( dblist, retname ):
  return getByName( dblist, retname, 'IDs' )

def retFileLoc( bstdout ):
  if bstdout == None:
    return None
  retstr = bstdout.decode('utf-8')
  if isWin():
    retstr = retstr.translate(str.maketrans({"\\": r"\\"}))
  fileloc = json.loads( retstr )
  if fileloc['Status'] != 'OK':
    log_msg( fileloc['Status'] )
    raise FileNotFoundError
  return fileloc['File_name']

def getFileLocation( dbkey, args=None ):
  cmd = getODCommand(dbmanexe,args)
  cmd.append( '--json' )
  cmd.append( '--info' )
  cmd.append( dbkey )
  return retFileLoc( execCommand(cmd) )

def getNewEntryFileName( objnm, dirid, trgrp, trl, ext, ftype=None, args=None ):
  cmd = getODCommand(dbmanexe,args)
  cmd.append( '--create' )
  cmd.append( objnm )
  cmd.append( dirid )
  cmd.append( trgrp )
  cmd.append( trl )
  cmd.append( ext )
  if ftype != None:
      cmd.append( ftype )
  cmd.append( '--json' )
  return retFileLoc( execCommand(cmd) )




#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : A. Huck
# DATE     : Nov 2018
#
# ranges tools
#

import collections
import math
import odpy.common
import numpy as np

def get_range( samp ):
  return range( samp[0], samp[1]+samp[2], samp[2] )

def arrayFromRange( rg ):
  if rg == None:
    return None
  return np.linspace( rg.start, rg.stop, len(rg), dtype=np.int32 )

def get_range_steps( samp ):
  return arrayFromRange( get_range(samp) )

def getLineObj( obj ):
  try:
    ret = obj['Inline']
  except (AttributeError, KeyError):
    try:
      ret = obj['lines']
    except (AttributeError, KeyError):
      ret = None
  return ret

def getTraceObj( obj ):
  try:
    ret = obj['Crossline']
  except (AttributeError, KeyError):
    try:
      ret = obj['traces']
    except (AttributeError, KeyError):
      ret = None
  return ret

def getZObj( obj ):
  try:
    ret = obj['Z']
  except (AttributeError, KeyError):
    try:
      ret = obj['zsamp']
    except (AttributeError, KeyError):
      ret = None
  return ret

def getAxesAsRanges( tkzs ):
   return collections.OrderedDict ({
    'lines': get_range( getLineObj(tkzs) ),
    'traces': get_range( getTraceObj(tkzs) ),
    'zsamp': get_range( getZObj(tkzs) )
   })

def getAxesAsArrays( tkzs ):
  return collections.OrderedDict ({
    'lines': get_range_steps( getLineObj(tkzs) ),
    'traces': get_range_steps( getTraceObj(tkzs) ),
    'zsamp': get_range_steps( getZObj(tkzs) )
  })

def getIntervalStr( rg, label ):
  if len(rg) < 3:
    return
  str = label + " range: "+ repr(rg[0]) + "-" + repr(rg[1])
  str = str + " (step " + repr(rg[2]) + ")"
  return str

def printSampling( tkzs ):
  print( getIntervalStr(getLineObj(tkzs),"Inline") )
  print( getIntervalStr(getTraceObj(tkzs),"Crossline") )
  print( getIntervalStr(getZObj(tkzs),"Z") )

def niceNumber(value, round=True):
  """Return a nicely rounded number.
    
    Parameters
    ----------
    value : float
      Input value.
    round : boolean, optional
      If true returned number will be rounded otherwise returns the ceiling. The default is True.
   
    Returns
    -------
    float
      The nice number.
   
    """
  if value > 0:
      signfact = 1
  else:
      signfact = -1
  value = np.abs(value)
  nicefrac = 0
  try:
      exponent = math.floor(math.log10(value))
  except ValueError:
      return value*signfact
  fraction = value/math.pow(10,exponent)
  if round:
    if fraction<1.5:
      nicefrac = 1
    elif fraction<3:
      nicefrac = 2
    elif fraction<7:
      nicefrac = 5
    else:
      nicefrac = 10
  else:
    if fraction<=1:
      nicefrac = 1
    elif fraction<=2:
      nicefrac = 2
    elif fraction<=5:
      nicefrac = 5
    else:
      nicefrac = 10
  return nicefrac * math.pow(10, exponent) * signfact
                    
def niceRange(min, max, maxsteps=10):
  """ Return a nicely rounded range - good for axes.
  
   Parameters
   ----------
   min : float
        Range minimum.
    max : float
        Range maximum.
    maxsteps : int, optional
        Maximum number of major tick steps desired in output range. The default is 10.

    Returns
    -------
    niceMin : float
       The nicely rounded minimum.
    niceMax : float
        The nicely rounded maximum.
    """
  range = abs(max-min)
  if math.isclose(min, max, rel_tol=1e-5):
      range = 0.2 * abs(min)
  range = niceNumber(range, False)
  tickspacing = niceNumber(range/(maxsteps), True)
  niceMin = math.floor(min/tickspacing)*tickspacing
  niceMax = math.ceil(max/tickspacing)*tickspacing
  return (niceMin, niceMax)



#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : A. Huck
# DATE     : June 2019
#
# Deep learning apply server
#
#

import argparse
import numpy as np
from os import path
import selectors
import socket
import sys
import time
import traceback

from odpy.common import *
from odpy import oscommand
import dgbpy.keystr as dgbkeys
import dgbpy.hdf5 as dgbhdf5
import dgbpy.deeplearning_apply_clientlib as applylib

sel = selectors.DefaultSelector()

# -- command line parser

parser = argparse.ArgumentParser(
          description='Client application of a trained machine learning model')
parser.add_argument( '-v', '--version',
            action='version',version='%(prog)s 2.0')
datagrp = parser.add_argument_group( 'Data' )
datagrp.add_argument( 'modelfile', type=argparse.FileType('r'),
                       help='The input trained model file' )
datagrp.add_argument( '--examplefile', 
                       dest='examples', metavar='file', nargs='?',
                       type=argparse.FileType('r'),
                       help='Examples file to be applied' )
netgrp = parser.add_argument_group( 'Network' )
netgrp.add_argument( '--address',
            dest='addr', metavar='ADDRESS', action='store',
            type=str, default='localhost',
            help='Address to listen on' )
netgrp.add_argument( '--port',
            dest='port', action='store',
            type=int, default=65432,
            help='Port to listen on')
loggrp = parser.add_argument_group( 'Logging' )
loggrp.add_argument( '--log',
            dest='logfile', metavar='file', nargs='?',
            type=argparse.FileType('w'), default=sys.stdout,
            help='Progress report output' )
loggrp.add_argument( '--syslog',
            dest='sysout', metavar='stdout', nargs='?',
            type=argparse.FileType('w'), default=sys.stdout,
            help='System log' )
loggrp.add_argument( '--server-log',
            dest='servlogfile', metavar='file', nargs='?',
            type=argparse.FileType('w'), default=sys.stdout,
            help='Python server log' )
loggrp.add_argument( '--server-syslog',
            dest='servsysout', metavar='stdout', nargs='?',
            type=argparse.FileType('w'), default=sys.stdout,
            help='Python server System log' )
# optional
parser.add_argument( '--fakeapply', dest='fakeapply', action='store_true',
                     default=False,
                     help="applies a numpy average instead of the model" )
parser.add_argument( '--local', dest='localserv', action='store_true',
                     default=False,
                     help="use a local network socket connection" )


args = vars(parser.parse_args())
initLogging( args )
modelfnm = args['modelfile'].name
local = args['localserv']

servscriptfp =  path.join(path.dirname(__file__),'deeplearning_apply-server.py')
servercmd = list()
servercmd.append( oscommand.getPythonExecNm() )
servercmd.append( servscriptfp )
servercmd.append( modelfnm )
servercmd.append( '--address' )
servercmd.append( str(args['addr']) )
servercmd.append( '--port' )
servercmd.append( str(args['port']) )
if args['servlogfile'].name != '<stdout>':
  servercmd.append( '--log' )
  servercmd.append( args['servlogfile'].name )
if args['servsysout'].name != '<stdout>':
  servercmd.append( '--syslog' )
  servercmd.append( args['servsysout'].name )
if args['fakeapply']:
  servercmd.append( '--fakeapply' )
if local:
  servercmd.append( '--local' )

serverproc = oscommand.execCommand( servercmd, True )
time.sleep( 2 )

def getApplyTrace( dict ):
  arr3d = dict['arr']
  shape = dict['inp_shape']
  procstep = dict['step']-1
  idx = dict['idx']
  idy = dict['idy']
  return arr3d[:,idx:idx+shape[0],\
                 idy:idy+shape[1]+procstep+1,:]

def create_request(action, value=None):
  if value == None:
    return dict(
      type="text/json",
      encoding="utf-8",
      content=dict(action=action),
    )
  elif action == 'outputs':
    return dict(
      type="text/json",
      encoding="utf-8",
      content=dict(
        action=action,
        value= {
          'names': value,
          dgbkeys.surveydictstr: 'None',
          dgbkeys.dtypepred: 'uint8',
          dgbkeys.dtypeprob: 'float32',
          dgbkeys.dtypeconf: 'float32'
        },
      ),
    )
  elif action == 'data':
    arr = getApplyTrace(value)
    return dict(
      type='binary/array',
      encoding=[arr.dtype.name],
      content=[arr],
    )
  else:
    return dict(
        type="binary/custom-client-binary-type",
        encoding="binary",
        content=bytes(action + value, encoding="utf-8"),
    )

def req_connection(host, port, request):
  if local:
    addr = str(port)
    sockfam = socket.AF_UNIX
  else:
    addr = (host, port)
    sockfam = socket.AF_INET
  sock = socket.socket(sockfam, socket.SOCK_STREAM)
  sock.setblocking(True)
  sock.connect_ex(addr)
  events = selectors.EVENT_READ | selectors.EVENT_WRITE
  message = applylib.Message(sel, sock, addr, request)
  sel.register(sock, events, data=message)

def getApplyPars( args ):
  if args['examples'] == None:
    ret= {
      'inp_shape': [33,33,33],
      'nrattribs': 1,
      'outputnms': dgbhdf5.getOutputNames(modelfnm,[0]),
      'surveydirnm': 'None'
    }
  else:
    exfnm = args['examples'].name
    info = dgbhdf5.getInfo( exfnm )
    shape = info['inp_shape']
    if not isinstance(shape,list):
      shape=[0,0,shape]
    ret = {
      'inp_shape': shape,
      'nrattribs': dgbhdf5.getNrAttribs( info ),
      'outputnms': dgbhdf5.getOutputs( exfnm )
    }
  return ret

pars = getApplyPars( args )
shape = pars['inp_shape']

nrattribs = pars['nrattribs']
nrlines_out = 20
nrtrcs_out = 800
chunk_step = 50
nrlines_in = nrlines_out + shape[0] - 1
nrtrcs_in = nrtrcs_out + shape[1] - 1
nrz_in = 378
nrz_out = nrz_in - shape[2] + 1
inpdata = np.random.random( nrattribs*nrlines_in*nrtrcs_in*nrz_in )
inpdata = inpdata.reshape((nrattribs,nrlines_in,nrtrcs_in,nrz_in))
inpdata = inpdata.astype('float32')

start = time.time()

host,port = args['addr'], args['port']
req_connection(host, port, create_request('status'))
req_connection(host, port, create_request('outputs',pars['outputnms']))
applydict = {
  'arr': inpdata,
  'inp_shape': shape,
  'idx': int(shape[0]/2-1),
  'idy': int(shape[1]/2-1),
  'step': chunk_step,
}
lastidy = nrtrcs_in-int(shape[1]/2-1)
nrrepeats = 1
trcrg = range(0,nrtrcs_in-shape[1]+1)
nrtrcs = (nrrepeats * len(trcrg))
for i in range(nrrepeats):
  applydict['idx'] = i
  for idy in range(0,nrtrcs_in-shape[1]+1,chunk_step):
    applydict['idy'] = idy
    req_connection(host, port, create_request('data',applydict))

req_connection(host, port, create_request('kill'))

try:
  while True:
    events = sel.select(timeout=1)
    for key, mask in events:
      message = key.data
      try:
        message.process_events(mask)
      except Exception:
        std_msg(
            "main: error: exception for",
            f"{message.addr}:\n{traceback.format_exc()}",
        )
        message.close()
    # Check for a socket being monitored to continue.
    if not sel.get_map():
      break
except KeyboardInterrupt:
  std_msg("caught keyboard interrupt, exiting")
finally:
  sel.close()
  oscommand.kill( serverproc )
  duration = time.time()-start
  log_msg( "Total time:",  "{:.3f}".format(duration), "s.;", \
         "{:.3f}".format(nrtrcs/duration), "tr/s." )



#__________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Apr 2019
#
# _________________________________________________________________________

from functools import partial

import numpy as np

from bokeh.layouts import column
from bokeh.models.widgets import CheckboxGroup, Div, Select, Slider

from odpy.common import log_msg
from dgbpy.dgbkeras import *
from dgbpy import uibokeh

info = None

def getPlatformNm( full=False ):
  if full:
    return platform
  return getMLPlatform()

def getSizeStr( sizeinbytes ):
  ret = 'Size: '
  try:
    import humanfriendly
    ret += humanfriendly.format_size( sizeinbytes )
  except Exception as e:
    ret += str(int(sizeinbytes)) + ' bytes'
  return ret

def getUiModelTypes( learntype, classification, ndim ):
  ret = ()
  for model in getModelsByType( learntype, classification, ndim ):
    ret += ((model,),)

  return dgbkeys.getNames( ret )

def getUiPars(uipars=None):
  dict = keras_dict
  learntype = info[dgbkeys.learntypedictstr]
  if isinstance(info[dgbkeys.inpshapedictstr], int):
    ndim = 1
  else:
    ndim = len(info[dgbkeys.inpshapedictstr])
  modeltypes = getUiModelTypes( learntype, info[dgbkeys.classdictstr], ndim )
  if len(modeltypes)==0:
    divfld = Div(text="""No Keras models found for this workflow.""")
    parsgrp = column(divfld)
    return {'grp': parsgrp,
            'uiobjects':{
              'divfld': divfld
            }
           }
  defmodel = modeltypes[0]
  defbatchsz = keras_dict['batch']
  estimatedsz = info[dgbkeys.estimatedsizedictstr]
  if kc.UserModel.isImg2Img( defmodel ):
      defbatchsz = 4
  uiobjs = {}
  if not uipars:
    uiobjs = {
      'modeltypfld': Select(title='Type', options=modeltypes),
      'batchfld': Select(title='Batch Size',options=cudacores),
      'epochfld': Slider(start=1,end=1000, title='Epochs'),
      'patiencefld': Slider(start=1,end=100, title='Patience'),
      'lrfld': Slider(start=-10,end=-1,step=1, title='Initial Learning Rate (1e)'),
      'edfld': Slider(start=1,end=100, title='Epoch drop (%)', step=0.1),
      'sizefld': None,
      'dodecimatefld': CheckboxGroup( labels=['Decimate input']),
      'chunkfld': Slider(start=1,end=100, title='Number of Chunks'),
      'rundevicefld': CheckboxGroup( labels=['Train on GPU'], visible=can_use_gpu())
    }
    if estimatedsz:
      uiobjs['sizefld'] = Div( text=getSizeStr( estimatedsz ) )
    uiobjs['dodecimatefld'].on_click(partial(decimateCB,chunkfld=uiobjs['chunkfld'],sizefld=uiobjs['sizefld']))
    try:
      uiobjs['chunkfld'].value_throttled = uiobjs['chunkfld'].value
      uiobjs['chunkfld'].on_change('value_throttled',partial(chunkfldCB, uiobjs['sizefld']))
    except AttributeError:
      log_msg( '[WARNING] Bokeh version too old, consider updating it.' )
      pass
    parsgrp = column(*list(uiobjs.values()))
    uipars = {'grp': parsgrp, 'uiobjects': uiobjs}
  else:
    uiobjs = uipars['uiobjects']
    
  uiobjs['modeltypfld'].value = defmodel
  uiobjs['batchfld'].value = str(defbatchsz)
  uiobjs['epochfld'].value = dict['epoch']
  uiobjs['patiencefld'].value = dict['patience']
  uiobjs['lrfld'].value = np.log10(dict['learnrate'])
  uiobjs['edfld'].value = 100*dict['epochdrop']/uiobjs['epochfld'].value
  if estimatedsz:
    uiobjs['sizefld'].text = getSizeStr(estimatedsz)
  uiobjs['dodecimatefld'].active = []
  uiobjs['chunkfld'].value = dict['nbchunk']
  uiobjs['rundevicefld'].active = [0]
  decimateCB( uiobjs['dodecimatefld'].active,uiobjs['chunkfld'],uiobjs['sizefld'] )
  return uipars

def chunkfldCB(sizefld,attr,old,new):
  if sizefld == None:
    return
  sizefld.text = getSizeStr( info[dgbkeys.estimatedsizedictstr]/new )

def getUiParams( keraspars ):
  kerasgrp = keraspars['uiobjects']
  nrepochs = kerasgrp['epochfld'].value
  epochdroprate = kerasgrp['edfld'].value / 100
  epochdrop = int(nrepochs*epochdroprate)
  if epochdrop < 1:
    epochdrop = 1
  runoncpu = not kerasgrp['rundevicefld'].visible or \
             not isSelected( kerasgrp['rundevicefld'] )
  return getParams( dodec=isSelected(kerasgrp['dodecimatefld']), \
                             nbchunk=kerasgrp['chunkfld'].value, \
                             epochs=kerasgrp['epochfld'].value, \
                             batch=int(kerasgrp['batchfld'].value), \
                             patience=kerasgrp['patiencefld'].value, \
                             learnrate= 10 ** kerasgrp['lrfld'].value, \
                             epochdrop=epochdrop, \
                             nntype=kerasgrp['modeltypfld'].value, \
                             prefercpu=runoncpu)

def isSelected( fldwidget, index=0 ):
  return uibokeh.integerListContains( fldwidget.active, index )

def decimateCB( widgetactivelist,chunkfld,sizefld ):
  decimate = uibokeh.integerListContains( widgetactivelist, 0 )
  chunkfld.visible = decimate
  if sizefld == None:
    return
  size = info[dgbkeys.estimatedsizedictstr]
  if decimate:
    size /= chunkfld.value
  sizefld.text = getSizeStr( size )



#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : A. Huck
# DATE     : Nov 2018
#
# tools for reading hdf5 files created by OpendTect
#

import collections
import numpy as np
import h5py

dGBInfoDSName = '++info++'

def openFile( filenm, mode ):
  return h5py.File( filenm, mode, libver=('v110', 'latest') )

def getInfoDataSet( h5file_or_grp ):
  try:
    return h5file_or_grp[dGBInfoDSName]
  except KeyError:
    print( "Input Error: HDF5 file probably not created by OpendTect (no ++info++ attribute)" )
    raise

def ensureHasDataset( h5file_or_grp ):
  if dGBInfoDSName in h5file_or_grp:
    return getInfoDataSet( h5file_or_grp )
  return h5file_or_grp.create_dataset(dGBInfoDSName,dtype='S1')

def getAttr( dataset, ky ):
  try:
    attrib = dataset.attrs[ky]
    return np.bytes_(dataset.attrs[ky]).decode().rstrip().split("\x00")[0]
  except KeyError:
    print( "HDF5 key not found: '"+ky+"'. Available keys:\n")
    print( list(dataset.attrs.keys()) )
    raise

def setAttr( dataset, ky, str ):
  dataset.attrs[ky] = np.chararray.encode( str )

def hasAttr( dataset, ky ):
  return ky in dataset.attrs

def getAttrs( dataset, ky ):
  return getAttr( dataset, ky ).split( "`" )

def getText( infos, ky ):
  ret = getAttr( infos, ky )
  if "`" in ret:
    return ret.split( "`" )
  return ret

def setArray( infos, ky, arr ):
  try:
    arrstr = '`'.join( map(str,arr) )
  except TypeError:
    arrstr = str(arr)
  setAttr( infos, ky, arrstr )

def getBoolValue( infos, ky ):
  return getAttr( infos, ky ) == "Yes"

def getIArray_( valsstr ):
  if isinstance(valsstr,list) and len(valsstr) > 1:
    ret = list()
    for valstr in valsstr:
      ret.append( int(valstr) )
    return ret
  try:
    return int(valsstr)
  except ValueError:
    return int(float(valsstr))

def getIntValue( infos, ky ):
  return getIArray_( getText(infos,ky) )

def getIInterval( infos, ky ):
  return getIArray_( getText(infos,ky) )

def getIStepInterval( infos, ky ):
  return getIArray_( getText(infos,ky) )

def getIArray( infos, ky ):
  return getIArray_( getText(infos,ky) )

def getDArray_( valsstr ):
  if isinstance(valsstr,list) and len(valsstr) > 1:
    ret = list()
    for valstr in valsstr:
      ret.append( float(valstr) )
    return ret
  return float(valsstr)
  
def getDValue( infos, ky ):
  return getDArray_( getText(infos,ky) )

def getDInterval( infos, ky ):
  return getDArray_( getText(infos,ky) )

def getDStepInterval( infos, ky ):
  return getDArray_( getText(infos,ky) )

def getDArray( infos, ky ):
  return getDArray_( getText(infos,ky) )

def getAttribInfo( filenm ):
  h5file = openFile( filenm, 'r' )
  infods = getInfoDataSet( h5file )

  blocksversion = getIntValue(infods,"Blocks.Version")
  zdomain  = getText(infods,"ZDomain")
  datasetnm = getText(infods,"Name.Cube")
  compnms  = getText(infods,"Components")
  datachar = getText(infods,"Data storage")
  inlinerg    = getIInterval(infods,"In-line range")
  crosslinerg = getIInterval(infods,"Cross-line range")
  zrg      = getDStepInterval(infods,"Z range")
  inl_step = getIntValue(infods,"Step In-line")
  crl_step = getIntValue(infods,"Step Cross-line")
  blocksdim = getIInterval(infods,"Blocks.Max Dimensions")
  blocksinlrg = getIInterval(infods,"Blocks.Inl ID Range")
  blockscrlrg = getIInterval(infods,"Blocks.Crl ID Range")
  h5file.close()
  if zdomain == "TWT":
    zrg = np.multiply(zrg,1000).astype("int32")
    zrg = [zrg[0],zrg[1],zrg[2]]
  return collections.OrderedDict({
    'name': datasetnm,
    'attributes': compnms,
    'version': blocksversion,
    'zdomain': zdomain,
    'storage': datachar,
    'range': collections.OrderedDict({
      'Inline': [inlinerg[0],inlinerg[1],inl_step],
      'Crossline': [crosslinerg[0],crosslinerg[1],crl_step],
      'Z': zrg
    }),
    'block': {
      'size': blocksdim,
      'range': collections.OrderedDict({
        'Inline': blocksinlrg,
        'Crossline': blockscrlrg
      })
    }
  })

# Survey transformation (X-Y) to (Inl-Crl)
# Given an Inl-Crl pair: ic_pos = [inl,crl]
# the following returns the corresponding X,Y location:
# xy_pos = transform['origin'] + np.dot( transform['rotation'], ic_pos )
def getTransform( info ):
  coordsxbid = getDStepInterval(info,"Coord-X-BinID")
  coordsybid = getDStepInterval(info,"Coord-Y-BinID")
  return collections.OrderedDict({
    'origin':   [coordsxbid[0],coordsybid[0]],
    'rotation': [[coordsxbid[1],coordsxbid[2]],[coordsybid[1],coordsybid[2]]]
  })

def getSurveyInfo( filenm ):
  h5file = openFile( filenm, 'r' )
  infods = getInfoDataSet( h5file )
  surveyname = getText(infods,"Name.Survey")
  inlstart = getIntValue(infods,"First In-line")
  inlstop  = getIntValue(infods,"Last In-line")
  crlstart = getIntValue(infods,"First Cross-line")
  crlstop  = getIntValue(infods,"Last Cross-line")
  xrg = getDInterval(infods,"X range")
  yrg = getDInterval(infods,"Y range")
  transform = getTransform( infods )
  h5file.close()
  return collections.OrderedDict({
    'name': surveyname,
    'range': collections.OrderedDict({
      'Inline': [inlstart,inlstop],
      'Crossline': [crlstart,crlstop],
      'X': xrg,
      'Y': yrg
    }),
    'transform': transform
  })



#__________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Apr 2019
#
# _________________________________________________________________________

from functools import partial

from bokeh.core import enums
from bokeh.layouts import row, column
from bokeh.models import Spacer
from bokeh.models.widgets import Button, Select, Slider

from odpy.common import log_msg
from dgbpy.dgbscikit import *
from dgbpy import uibokeh

but_width = uibokeh.but_width
but_height = uibokeh.but_height
but_spacer = uibokeh.but_spacer
defaultbut = enums.ButtonType.default

def getPlatformNm( full=False ):
  if full:
    return platform
  return getMLPlatform()

def getLinearGrp(uipars=None):
  uiobjs = {}
  if not uipars:
    uiobjs = {'lineartyp': Select(title='Model', options=getUiLinearTypes()),}
    uipars = {'uiobjects': uiobjs, 'name': regmltypes[0][1],}
  else:
    uiobjs = uipars['uiobjects']
    
  uiobjs['lineartyp'].value = 'Ordinary Least Squares'
  return uipars

def getLogGrp(uipars=None):
  uiobjs = {}
  if not uipars:
    uiobjs = {'logtyp': Select(title='Model', options=getUiLogTypes()),
              'solvertyp': Select(title='Solver', options=getUiSolverTypes())
             }
    uipars = {'uiobjects': uiobjs, 'name': classmltypes[0][1],}
  else:
    uiobjs = uipars['uiobjects']
    
  uiobjs['logtyp'].value = 'Logistic Regression Classifier'
  uiobjs['solvertyp'].value = getDefaultSolver()
  return uipars

def getEnsembleGrp(uipars=None):
  uiobjs = {}
  rfgrp = {}
  xgdtgrp = None
  xgrfgrp = None
  if uipars:
    uiobjs  = uipars['uiobjects']
    if hasXGBoost():
      xgdtgrp = {'uiobjects': {'depparfldxgdt': uiobjs['depparfldxgdt'],
                               'estparfldxgdt': uiobjs['estparfldxgdt'],
                               'lrparfldxgdt': uiobjs['lrparfldxgdt'],}, }
      xgrfgrp = {'uiobjects': {'depparfldxgrf': uiobjs['depparfldxgrf'],
                               'estparfldxgrf': uiobjs['estparfldxgrf'],
                               'lrparfldxgrf': uiobjs['lrparfldxgrf'],}, }
    rfgrp = {'uiobjects': {'depparfldrf': uiobjs['depparfldrf'],
                           'estparfldrf': uiobjs['estparfldrf'],},}
    gbgrp = {'uiobjects': {'depparfldgb': uiobjs['depparfldgb'],
                           'estparfldgb': uiobjs['estparfldgb'], 
                           'lrparfldgb': uiobjs['lrparfldgb'],}, }
    adagrp = {'uiobjects': {'estparfldada': uiobjs['estparfldada'],
                            'lrparfldada': uiobjs['lrparfldada'],}, }
    if hasXGBoost():
      getXGDTGrp(xgdtgrp)
      getXGRFGrp(xgrfgrp)
    getRFGrp(rfgrp)
    getGBGrp(gbgrp)
    getAdaGrp(adagrp)
  else:
    if hasXGBoost():
      xgdtgrp = getXGDTGrp()
      xgrfgrp = getXGRFGrp()
    rfgrp = getRFGrp()
    gbgrp = getGBGrp()
    adagrp = getAdaGrp()
    uiobjs = {'ensembletyp': Select(title='Model',
                             options=getUiEnsembleTypes())}
    if hasXGBoost():
      xggrpuiobjs = xgdtgrp['uiobjects']
      uiobjs.update({
                      'depparfldxgdt': xggrpuiobjs['depparfldxgdt'],
                      'estparfldxgdt': xggrpuiobjs['estparfldxgdt'],
                      'lrparfldxgdt': xggrpuiobjs['lrparfldxgdt'],
                    })
      xggrpuiobjs = xgrfgrp['uiobjects']
      uiobjs.update({
                      'depparfldxgrf': xggrpuiobjs['depparfldxgrf'],
                      'estparfldxgrf': xggrpuiobjs['estparfldxgrf'],
                      'lrparfldxgrf': xggrpuiobjs['lrparfldxgrf'],
                    })
    uiobjs.update({
              'depparfldrf': rfgrp['uiobjects']['depparfldrf'],
              'estparfldrf': rfgrp['uiobjects']['estparfldrf'],
              'depparfldgb': gbgrp['uiobjects']['depparfldgb'],
              'estparfldgb': gbgrp['uiobjects']['estparfldgb'], 
              'lrparfldgb': gbgrp['uiobjects']['lrparfldgb'],
              'estparfldada': adagrp['uiobjects']['estparfldada'],
              'lrparfldada': adagrp['uiobjects']['lrparfldada'],
              })
    if hasXGBoost():
      ensemblegrp = (xgdtgrp,xgrfgrp)
    ensemblegrp += (rfgrp,gbgrp,adagrp,)
    uiobjs['ensembletyp'].on_change('value',partial(ensembleChgCB,cb=uiobjs['ensembletyp'],ensemblegrp=ensemblegrp))
    uipars = {'uiobjects': uiobjs, 'name': regmltypes[1][1],}
    
  if hasXGBoost():
    uiobjs['ensembletyp'].value = 'XGBoost: (Decision Tree)'
  else:
    uiobjs['ensembletyp'].value = 'Random Forests'
  return uipars

def getNNGrp(uipars=None):
  dict = scikit_dict
  uiobjs = {}
  if not uipars:
    uiobjs = {'nntyp': Select(title='Model', options=getUiNNTypes()),
              'itrparfld': Slider(start=10,end=1000, title='Max Iteration'),
              'lay1parfld': Slider(start=1,end=200, title='Layer 1'),
              'lay2parfld': Slider(start=1,end=50, title='Layer 2'),
              'lay3parfld': Slider(start=1,end=10, title='Layer 3'),
              'lay4parfld': Slider(start=1,end=5, title='Layer 4'),
              'lay5parfld': Slider(start=1,end=3, title='Layer 5'),
              'lrparfld': Slider(start=1,end=100, title='Initial Learning Rate '+ '('+u'\u2030'+')'),
              'addbutton': Button(label='Add',button_type=defaultbut,width=but_width,height=but_height),
              'lessbutton': Button(label='Less',button_type=defaultbut,width=but_width,height=but_height),
              }
    uiobjs['buttonparfld'] = row(uiobjs['addbutton'],Spacer(width = 5),uiobjs['lessbutton'],sizing_mode='stretch_width')
    nb = Slider(start=1,end=5,value=3)
    layergrp = [nb,uiobjs['lay1parfld'],uiobjs['lay2parfld'],uiobjs['lay3parfld'],
                uiobjs['lay4parfld'],uiobjs['lay5parfld'],uiobjs['addbutton'],uiobjs['lessbutton']]
    try:
      uiobjs['lay1parfld'].value_throttled = uiobjs['lay1parfld'].value
      uiobjs['lay2parfld'].value_throttled = uiobjs['lay2parfld'].value
      uiobjs['lay3parfld'].value_throttled = uiobjs['lay3parfld'].value
      uiobjs['lay4parfld'].value_throttled = uiobjs['lay4parfld'].value
      uiobjs['lay1parfld'].on_change('value_throttled',partial(layer1ChgCB,layergrp))
      uiobjs['lay2parfld'].on_change('value_throttled',partial(layer2ChgCB,layergrp))
      uiobjs['lay3parfld'].on_change('value_throttled',partial(layer3ChgCB,layergrp))
      uiobjs['lay4parfld'].on_change('value_throttled',partial(layer4ChgCB,layergrp))
    except AttributeError:
      log_msg( '[WARNING] Bokeh version too old, consider updating it.' )
      pass
    uiobjs['addbutton'].on_click(partial(buttonChgCB,uiobjs['addbutton'],layergrp))
    uiobjs['lessbutton'].on_click(partial(buttonChgCB,uiobjs['lessbutton'],layergrp))
    uipars = {'uiobjects': uiobjs,     
              'name': regmltypes[2][1],
              'nb': nb
             }
  else:
    uiobjs = uipars['uiobjects']
    
  uiobjs['nntyp'].value = 'Multi-Layer Perceptron'
  uiobjs['itrparfld'].value = dict['nnpars']['maxitr']
  uiobjs['lay1parfld'].value = 50
  uiobjs['lay2parfld'].value = dict['nnpars']['lay2']
  uiobjs['lay3parfld'].value = dict['nnpars']['lay3']
  uiobjs['lay4parfld'].value = dict['nnpars']['lay4']
  uiobjs['lay5parfld'].value = dict['nnpars']['lay5']
  uiobjs['lrparfld'].value = dict['nnpars']['lr']*1000
  uipars['nb'].value = 3
  return uipars

def getSVMGrp(uipars=None):
  dict = scikit_dict
  uiobjs = {}
  defkernelstr = getDefaultNNKernel()
  if not uipars:
    uiobjs = {'svmtyp': Select(title='Model', options=getUiSVMTypes()),
              'kernel': Select(title='Kernel', options=getUiNNKernelTypes()),
              'degree': Slider(start=1,end=5, step=1,title='Degree'),
              }
    uiobjs['kernel'].on_change('value',partial(kernelChgCB,deg=uiobjs['degree']))
    kernelChgCB( 'value', defkernelstr, defkernelstr, uiobjs['degree'])
    uipars = {'uiobjects': uiobjs, 'name': regmltypes[3][1],}
  else:
    uiobjs = uipars['uiobjects']
    
  uiobjs['svmtyp'].value = 'Support Vector Machine'
  uiobjs['kernel'].value = defkernelstr
  uiobjs['degree'].value = dict['svmpars']['degree']
  return uipars

#deg visible only when polynomial kernel function
def kernelChgCB( attrnm, old, new,deg):
  if new == 'Polynomial':
    deg.visible = True
  else:
    deg.visible = False

def layer1ChgCB(layergrp,attr,old,new):
  layergrp[2].end = new
  if new <= layergrp[2].value:
    layergrp[2].value = new
    layer2ChgCB(layergrp,attr,layergrp[2].value,new)

def layer2ChgCB(layergrp,attr,old,new):
  layergrp[3].end = new
  if new <= layergrp[3].value:
    layergrp[3].value = new
    layer3ChgCB(layergrp,attr,layergrp[2].value,new)
  
def layer3ChgCB(layergrp,attr,old,new):
  layergrp[4].end = new
  if new <= layergrp[4].value:
    layergrp[4].value = new
    layer4ChgCB(layergrp,attr,layergrp[2].value,new)
  
def layer4ChgCB(layergrp,attr,old,new):
  layergrp[5].end = new
  if new <= layergrp[5].value:
    layergrp[5].value = new

def buttonChgCB(addbutton,layergrp):
  log_msg('Working')
  nb = layergrp[0].value #number of layers
  if addbutton.label == 'Add':
    layergrp[nb+1].visible = True
    if nb == 4:
      layergrp[6].visible = False # Hide add button
    elif nb == 1:
      layergrp[7].visible = True # Show less button
    layergrp[0].value += 1
  elif addbutton.label == 'Less':
    layergrp[nb].visible = False
    if nb == 5:
      layergrp[6].visible = True # show add button
    elif nb == 2:
      layergrp[7].visible = False # Hide less button
    layergrp[0].value -= 1
  else:
    return None

def modelChgCB( attrnm, old, new, cb, modelsgrp ):
  ret = uibokeh.getAllUiFlds( modelsgrp )
  for uifld in  ret:
    uifld.visible = False
  ret = uibokeh.getGroup( new, cb.options, modelsgrp, 'uiobjects' )
  # set default visibility
  if new == 'Ensemble':
    ret['ensembletyp'].visible = True
    ret['estparfldxgdt'].visible = True
    ret['depparfldxgdt'].visible = True
    ret['lrparfldxgdt'].visible = True
  elif new == 'SVM':
    ret['svmtyp'].visible = True
    ret['kernel'].visible = True
  else:
    for uifld in  ret:
      ret[uifld].visible = True
    if new == 'Neural Network':
      ret['lay4parfld'].visible = False
      ret['lay5parfld'].visible = False

def getXGDTGrp(uipars=None):
  if not hasXGBoost():
    return None
  dict = scikit_dict['ensemblepars']['xgdt']
  uiobjs = {}
  if not uipars:
    uiobjs = {'estparfldxgdt': Slider(start=10,end=1000,step=10,title='Estimators'),
              'depparfldxgdt': Slider(start=10,end=200,step=10,title='Max Depth'),
              'lrparfldxgdt': Slider(start=0.1,end=10,step=0.1,title='Learning Rate'), }
    uipars = {'uiobjects': uiobjs,}
  else:
    uiobjs = uipars['uiobjects']
    
  uiobjs['estparfldxgdt'].value = dict['est']
  uiobjs['depparfldxgdt'].value = dict['maxdep']
  uiobjs['lrparfldxgdt'].value = dict['lr']
  return uipars

def getXGRFGrp(uipars=None):
  if not hasXGBoost():
    return None
  dict = scikit_dict['ensemblepars']['xgrf']
  uiobjs = {}
  if not uipars:
    uiobjs = {'estparfldxgrf': Slider(start=10,end=1000,step=10,title='Estimators'),
              'depparfldxgrf': Slider(start=10,end=200,step=10,title='Max Depth'),
              'lrparfldxgrf': Slider(start=0.1,end=10,step=0.1,title='Learning Rate'), }
    uipars = {'uiobjects': uiobjs,}
  else:
    uiobjs = uipars['uiobjects']
    
  uiobjs['estparfldxgrf'].value = dict['est']
  uiobjs['depparfldxgrf'].value = dict['maxdep']
  uiobjs['lrparfldxgrf'].value = dict['lr']
  return uipars

def getRFGrp(uipars=None):
  dict = scikit_dict['ensemblepars']['rf']
  uiobjs = {}
  if not uipars:
    uiobjs = {'estparfldrf': Slider(start=10, end=1000, step=10, title='Estimators'),
              'depparfldrf': Slider(start=10, end=200, step=10, title='Max Depth'), }
    uipars = {'uiobjects': uiobjs}
  else:
    uiobjs = uipars['uiobjects']
    
  uiobjs['estparfldrf'].value = dict['est']
  uiobjs['depparfldrf'].value = dict['maxdep']
  return uipars

def getGBGrp(uipars=None):
  dict = scikit_dict['ensemblepars']['gb']
  uiobjs = {}
  if not uipars:
    uiobjs = {'estparfldgb': Slider(start=10,end=500,step=10,title='Estimators'),
              'depparfldgb': Slider(start=1,end=100,step=1,title='Max Depth'),
              'lrparfldgb': Slider(start=0.1,end=10,step=0.1,title='Learning Rate'), }
    uipars = {'uiobjects': uiobjs,}
  else:
    uiobjs = uipars['uiobjects']
    
  uiobjs['estparfldgb'].value = dict['est']
  uiobjs['depparfldgb'].value = dict['maxdep']
  uiobjs['lrparfldgb'].value = dict['lr']
  return uipars

def getAdaGrp(uipars=None):
  dict = scikit_dict['ensemblepars']['ada']
  uiobjs = {}
  if not uipars:
    uiobjs = {'estparfldada': Slider(start=10,end=500,step=10,title='Estimators'),
              'lrparfldada': Slider(start=0.1,end=10,step=0.1,title='Learning Rate'),
              }
    uipars = {'uiobjects': uiobjs,}
  else:
    uiobjs = uipars['uiobjects']
    
  uiobjs['estparfldada'].value = dict['est']
  uiobjs['lrparfldada'].value = dict['lr']
  return uipars

def ensembleChgCB( attrnm, old, new, cb, ensemblegrp ):
  ret = uibokeh.getAllUiFlds( ensemblegrp )
  for uifld in  ret:
    uifld.visible = False
  ret = uibokeh.getGroup( new, cb.options, ensemblegrp, 'uiobjects' )
  for uifld in  ret:
    ret[uifld].visible = True

def getUiPars(isclassification, uipars=None):
  models = getUiModelTypes(isclassification)
  uiobjs = {}
  parsgrp = None
  linearkey = 'loggrp' if isclassification else 'lineargrp'
  deftype = classmltypes[0][1] if isclassification else regmltypes[1][1]
  
  if not uipars:
    uiobjs = {
      'modeltyp': Select(title='Type',options=models),
      linearkey: getLogGrp() if isclassification else getLinearGrp(),
      'ensemblegrp': getEnsembleGrp(),
      'nngrp': getNNGrp(),
      'svmgrp': getSVMGrp()
    }
    modelsgrp = (uiobjs[linearkey], uiobjs['ensemblegrp'], uiobjs['nngrp'], uiobjs['svmgrp'])
    uiobjs['modeltyp'].on_change('value',partial(modelChgCB,cb=uiobjs['modeltyp'],modelsgrp=modelsgrp))
    pars = [uiobjs['modeltyp']]
    ensemblepars = [uiobjs['ensemblegrp']['uiobjects']['ensembletyp']]
    if hasXGBoost():
      xgdtpars = [uiobjs['ensemblegrp']['uiobjects']['estparfldxgdt'], \
                  uiobjs['ensemblegrp']['uiobjects']['depparfldxgdt'], \
                  uiobjs['ensemblegrp']['uiobjects']['lrparfldxgdt'], \
                 ]
      xgrfpars = [uiobjs['ensemblegrp']['uiobjects']['estparfldxgrf'], \
                  uiobjs['ensemblegrp']['uiobjects']['depparfldxgrf'], \
                  uiobjs['ensemblegrp']['uiobjects']['lrparfldxgrf'], \
                 ]
      ensemblepars.extend( xgdtpars )
      ensemblepars.extend( xgrfpars )
    ensemblepars.extend([
                    uiobjs['ensemblegrp']['uiobjects']['estparfldrf'], \
                    uiobjs['ensemblegrp']['uiobjects']['depparfldrf'], \
                    uiobjs['ensemblegrp']['uiobjects']['estparfldgb'], \
                    uiobjs['ensemblegrp']['uiobjects']['depparfldgb'], \
                    uiobjs['ensemblegrp']['uiobjects']['lrparfldgb'], \
                    uiobjs['ensemblegrp']['uiobjects']['estparfldada'], \
                    uiobjs['ensemblegrp']['uiobjects']['lrparfldada'], \
                   ] )
    nnpars = [uiobjs['nngrp']['uiobjects']['nntyp'], \
              uiobjs['nngrp']['uiobjects']['itrparfld'], \
              uiobjs['nngrp']['uiobjects']['lrparfld'], \
              uiobjs['nngrp']['uiobjects']['lay1parfld'], \
              uiobjs['nngrp']['uiobjects']['lay2parfld'], \
              uiobjs['nngrp']['uiobjects']['lay3parfld'], \
              uiobjs['nngrp']['uiobjects']['lay4parfld'], \
              uiobjs['nngrp']['uiobjects']['lay5parfld'], \
              uiobjs['nngrp']['uiobjects']['buttonparfld'], \
             ]
    svmpars = [uiobjs['svmgrp']['uiobjects']['svmtyp'], \
               uiobjs['svmgrp']['uiobjects']['kernel'], \
               uiobjs['svmgrp']['uiobjects']['degree'] \
              ]

    if isclassification:
      pars.extend([uiobjs[linearkey]['uiobjects']['logtyp'],
                   uiobjs[linearkey]['uiobjects']['solvertyp']])
    else:
      pars.extend([uiobjs[linearkey]['uiobjects']['lineartyp']])
    pars.extend(ensemblepars)
    pars.extend(nnpars)
    pars.extend(svmpars)
    parsgrp = column(*pars)
    uipars = {'grp': parsgrp, 'uiobjects': uiobjs}
  else:
    uiobjs = uipars['uiobjects']
    
  if isclassification:
    uiobjs['modeltyp'].value = models[0]
    uiobjs[linearkey] = getLogGrp(uiobjs[linearkey])
  else:
    uiobjs['modeltyp'].value = models[1]
    uiobjs[linearkey] = getLinearGrp(uiobjs[linearkey])
  uiobjs['ensemblegrp'] = getEnsembleGrp(uiobjs['ensemblegrp'])
  uiobjs['nngrp'] = getNNGrp(uiobjs['nngrp'])
  uiobjs['svmgrp'] = getSVMGrp(uiobjs['svmgrp'])

  modelsgrp = (uiobjs[linearkey], uiobjs['ensemblegrp'], uiobjs['nngrp'], uiobjs['svmgrp'])
  modelChgCB( 'value', deftype, deftype, uiobjs['modeltyp'], modelsgrp )
  return uipars

def getUiParams( sklearnpars ):
  sklearngrp = sklearnpars['uiobjects']
  modeltype = sklearngrp['modeltyp']
  if modeltype.value == 'Linear':
    parmobj = sklearngrp['lineargrp']['uiobjects']
    return getLinearPars( modelname=parmobj['lineartyp'].value )
  if modeltype.value == 'Logistic':
    parmobj = sklearngrp['loggrp']['uiobjects']
    return getLogPars( modelname=parmobj['logtyp'].value,
                       solver=parmobj['solvertyp'].value)
  if modeltype.value == 'Ensemble':
    parmobj = sklearngrp['ensemblegrp']['uiobjects']
    if parmobj['ensembletyp'].value == 'XGBoost: (Decision Tree)':
      return getEnsembleParsXGDT( modelname=parmobj['ensembletyp'].value,
                                  maxdep=parmobj['depparfldxgdt'].value,
                                  est=parmobj['estparfldxgdt'].value,
                                  lr=parmobj['lrparfldxgdt'].value )
    elif parmobj['ensembletyp'].value == 'XGBoost: (Random Forests)':
      return getEnsembleParsXGRF( modelname=parmobj['ensembletyp'].value,
                                  maxdep=parmobj['depparfldxgrf'].value,
                                  est=parmobj['estparfldxgrf'].value,
                                  lr=parmobj['lrparfldxgrf'].value )
    elif parmobj['ensembletyp'].value == 'Random Forests':
      return getEnsembleParsRF( modelname=parmobj['ensembletyp'].value,
                                maxdep=parmobj['depparfldrf'].value,
                                est=parmobj['estparfldrf'].value)
    elif parmobj['ensembletyp'].value == 'Gradient Boosting':
      return getEnsembleParsGB( modelname=parmobj['ensembletyp'].value,
                                maxdep=parmobj['depparfldgb'].value,
                                est=parmobj['estparfldgb'].value,
                                lr=parmobj['lrparfldgb'].value)
    elif parmobj['ensembletyp'].value == 'Adaboost':
      return getEnsembleParsAda( modelname=parmobj['ensembletyp'].value,
                                 est=parmobj['estparfldada'].value,
                                 lr=parmobj['lrparfldada'].value)
  elif modeltype.value == 'Neural Network':
    parmobj = sklearngrp['nngrp']['uiobjects']
    return getNNPars( modelname=parmobj['nntyp'].value,
                      maxitr=parmobj['itrparfld'].value,
                      lr=parmobj['lrparfld'].value/1000,
                      lay1=parmobj['lay1parfld'].value,
                      lay2=parmobj['lay2parfld'].value,
                      lay3=parmobj['lay3parfld'].value,
                      lay4=parmobj['lay4parfld'].value,
                      lay5=parmobj['lay5parfld'].value,
                      nb=sklearngrp['nngrp']['nb'].value)
  elif modeltype.value == 'SVM':
    parmobj = sklearngrp['svmgrp']['uiobjects']
    return getSVMPars( modelname=parmobj['svmtyp'].value,
                       kernel=parmobj['kernel'].value,
                       degree=parmobj['degree'].value )
  return None



"""
============================================================
Log plotting using Bokeh
============================================================

 Author:    Wayne Mogg
 Copyright: dGB Beheer BV
 Date:      June 2020


"""
import odpy.wellman as odwm
from bokeh.events import ButtonClick, Event
import bokeh.plotting as bp
import bokeh.models as bm
import bokeh.layouts as bl
import bokeh.core.enums as bce
from bokeh.core.properties import DashPattern
import math
import pandas as pd
import numpy as np
import random
from odpy.ranges import niceRange, niceNumber

class Well:
    def __init__(self,wellname):
        self.wellname = wellname
        self.track = None
        self.markers = None
        self.logcache = None
        self.logdata = None
        self.logdataview = None
        self.logdataviewidx = None
        self.limits = {}

    def getLogNames(self):
        if not self.logdata:
            return odwm.getLogNames(self.wellname)
        else:
            return self.logdata.column_names[1:]

    def getLogsFromFile(self, filenm, undefval=1e30):
        ld = pd.read_csv(filenm, delimiter='\t')
        ld = ld.replace(to_replace=undefval, value=float('Nan'))
        self.logdata = bm.ColumnDataSource(ld)
        self.logdataviewidx = self.logdata.data['index']
        self.logdataview = bm.CDSView(source=self.logdata, filters=[])
        return self.logdata

    def getLogLimits(self, lognm):
        limits = self.limits.get(lognm)
        if not limits:
            log = None
            if not self.logdata:
                log = self.getLog(lognm).data.get(lognm)
            else:
                log = self.logdata.data.get(lognm)
            limits = niceRange(np.nanmin(log), np.nanmax(log))
            self.limits[lognm] = limits

        return limits

    def setDepthView(self, mindepth, maxdepth):
        if self.logdataview:
            self.logdataviewidx = [idx for idx, z in enumerate(self.logdata.data['MD']) if z>=mindepth and z<=maxdepth]
            self.logdataview.filters = [bm.IndexFilter(indices=self.logdataviewidx)]

    def setLogLimits(self, lognm, left, right):
        self.limits[lognm] = [left, right]

    def getLog(self, lognm):
        if (not self.logcache) or (lognm not in self.logcache):
            (depths,logvals) = odwm.getLog(self.wellname, lognm)
            if not self.logcache:
               self.logcache = {}
            self.logcache[lognm] = bm.ColumnDataSource({'depth': depths, lognm: logvals})
        return self.logcache[lognm]

    def getTrack(self):
        if not self.track:
            self.track = odwm.getTrack(self.wellname)
        return self.track

    def getMarkers(self):
        if not self.markers:
            (mrkrs, depths, colors) = odwm.getMarkers(self.wellname)
            self.markers = bm.ColumnDataSource({'name': mrkrs, 'depth': depths, 'color': colors})
        return self.markers

    def depthRange(self):
        if self.logdata:
            return self.getLogLimits('MD')
        elif self.logcache:
            depthrange = None
            first = True
            for (key, item) in self.logcache.items():
                zlog = item.data['depth']
                if first:
                    first = False
                    depthrange = [zlog[0], zlog[-1]]
                else:
                    depthrange = [min(zlog[0],depthrange[0]), max(zlog[-1],depthrange[-1])]

            return depthrange
        else:
            return self.getTrack()[0]

class LogTrackMgr:
    def __init__(self, well, deflogs=None, trackwidth=400, withui=False):
        self.well = well
        self.width = trackwidth
        self.withui = withui
        self.tracks = {}
        self.tracklayout = None
        self.trackid = 0
        lognames = self.well.getLogNames()
        if not deflogs:
            self.deflogs = [lognames[0]]
        else:
            self.deflogs = list(set.intersection(set(deflogs), set(lognames)))
        self._addtrack()
        tracktabs = self.tracklayout.children[0].children[0]
        tracktabs.active = 1
        if len(deflogs)<=1 :
          tracktabs.active = 0


    def _addtrack(self, copytrack=None):

        if self.withui:
            addbutton = bm.Button(label='+', button_type='success', width_policy='min')
            addbutton.on_event(ButtonClick, self._addbuttonCB)
            rmvbutton = bm.Button(label='-', button_type='warning', width_policy='min')
            rmvbutton.on_event(ButtonClick, self._rmvbuttonCB)
            syncbutton = bm.CheckboxGroup(labels=['Sync'], active=[0], width_policy='min')
            syncbutton.on_change('active', self._syncbuttonCB)
            newtrack = LogTrack(self.well, self.width, withui=self.withui)
            if copytrack:
                newtrack.log_props = copytrack.log_props.copy()
                newtrack.track_props = copytrack.track_props.copy()
                master_trackbox = self.tracklayout.children[0]
                master_trackfig = self.tracks[master_trackbox.name].track
                newtrack.track.y_range.start = master_trackfig.y_range.start
                newtrack.track.y_range.end = master_trackfig.y_range.end
                newtrack.track.y_range.bounds = "auto"

                master_trackfig.y_range.js_on_change('start',
                          bm.CustomJS(args=dict(rg=newtrack.track.y_range,sb=syncbutton),
                                      code="if (sb.active.length===1) rg.start = this.start;"))
                master_trackfig.y_range.js_on_change('end',
                          bm.CustomJS(args=dict(rg=newtrack.track.y_range,sb=syncbutton),
                                      code="if (sb.active.length===1) rg.end = this.end;"))
                newtrack.apply_track_props()
                logs = copytrack.log_select.value
                newtrack.addLogs(logs)
                markers = copytrack.marker_select.value
                newtrack.addMarkers(markers)
            else:
                newtrack.addLogs(self.deflogs)
                rmvbutton.visible = False
                syncbutton.visible = False

            buttons = bl.row([addbutton,rmvbutton,syncbutton], align='center')
            trackbox = bl.column([newtrack.display(), buttons], name=str(self.trackid), tags=[addbutton.id, rmvbutton.id])
            trackbox.children[0].active = 1
            self.tracks[str(self.trackid)] = newtrack
            self.trackid += 1
            if self.tracklayout:
                self.tracklayout.children.append(trackbox)
                self._syncbuttonCB('active', 0, 0)
            else:
                self.tracklayout = bl.row([trackbox], sizing_mode='stretch_height')
                rmvbutton.visible = False
                syncbutton.visible = False
        else:
            self.tracklayout = LogTrack(self.well, self.width, withui=False)

    def _removetrack(self):
        pass

    def _addbuttonCB(self, event):
        evid = event._model_id
        for trackbox in self.tracklayout.children:
            if evid in trackbox.tags:
                self._addtrack(copytrack=self.tracks[trackbox.name])
                break


    def _rmvbuttonCB(self, event ):
        evid = event._model_id
        for idx in range(1, len(self.tracklayout.children)):
            trackbox = self.tracklayout.children[idx]
            if evid in trackbox.tags:
                self.tracks.pop(trackbox.name)
                self.tracklayout.children.pop(idx)
                break

    def _syncbuttonCB(self, attr, old, new):
        numtracks = len(self.tracklayout.children)
        for idx in range(1, numtracks):
          trackbox = self.tracklayout.children[idx]
          trackid = trackbox.name
          syncbutton = trackbox.children[-1].children[-1]
          logtrack = self.tracks[trackid]
          trackfig = logtrack.track
          if len(syncbutton.active)>0 and idx>0:
            trackfig.tools = []
          else:
            trackfig.tools = logtrack.default_tools
            trackfig.toolbar.active_scroll = logtrack.active_scroll



class LinePropertyWidget:
    def __init__(self):
        self.fields = {'width': bm.Spinner(title='Line Width:',low=0,high=5, step=1,
                                           width_policy='min'),
                       'color': bm.ColorPicker(title='Line Color:', width_policy='min'),
                       'dash':  bm.Select(title='Line Draw Style:', options=list(bce.DashPattern),
                                          width_policy='min')
                       }
        self.layout = bl.row(list(self.fields.values()))

class GridPropertyWidget:
    def __init__(self, title, withsteps=True):
        line_props = LinePropertyWidget()
        field_title = bm.Div(text=title)
        self.fields = {'visible': bm.CheckboxGroup(labels=[''], inline= True, active=[0], width_policy='min'),
                       'steps': bm.Spinner(title='Steps:',low=0,high=5, step=1)
                      }
        self.fields.update(line_props.fields)
        self.fields['visible'].on_change('active',self.visible)
        self.layout = bl.column(bl.row(self.fields['visible'],field_title),
                                self.fields['steps'],
                                line_props.layout
                                )

    def visible(self, attr, old, new):
        disable = False
        if len(new)==0:
            disable = True
        for (key, field) in self.fields.items():
            if key is 'visible':
                continue
            field.disabled = disable

class LogTrack:
    def __init__(self, well, width, withui=False):
        self.well = well
        self.width = width
        self.withui = withui
        self.tabs = None
        self.tracklayout = None
        self.tooltips = [("Depth", " $y"),
                         ("Log","$x"),
                        ]
        self.track_props = {}
        self.track_fields = {}
        self._inittrackprops()
        self.log_fields = {}
        self.log_props = {}
        self._initlogprops()
#        self.shading =np.empty((height,width,4),dtype=np.uint8)
        self._inittrack()
        self._initui()
        self.default_tools = self.track.tools
        self.active_scroll = self.track.toolbar.active_scroll

    def _inittrack(self):
        depths = self.well.depthRange()
        self.track = bp.figure(title=self.well.wellname,
                             plot_width=self.track_props['plot_width'],
                             sizing_mode='stretch_height',
                             background_fill_color=self.track_props['background_fill_color'],
                             tools='ypan, reset, hover',
                             y_axis_label='Depth (mMD)',
                             tooltips=self.tooltips)
        wheelzoom = bm.WheelZoomTool(dimensions="height", maintain_focus=False)
        self.track.add_tools(wheelzoom)
        boxzoom = bm.BoxZoomTool(dimensions="height")
        self.track.add_tools(boxzoom)
        self.track.toolbar.active_scroll = wheelzoom
        self.track.toolbar.logo = None
        self.track.title.text_font_size = '14pt'
        self.track.y_range = bm.Range1d(depths[-1], depths[0],bounds="auto")
        self.track.x_range = bm.Range1d(0, 10)
        self.track.xaxis.visible = False
        self.apply_track_props()


    def _inittrackprops(self):
        tmp = bp.figure()
        self.track_props = {'plot_width': self.width,
                            'background_fill_color': tmp.background_fill_color,
                            'log_major_visible': tmp.xgrid.visible,
                            'log_major_num': tmp.xaxis[0].ticker.desired_num_ticks-1,
                            'log_major_width': tmp.xgrid.grid_line_width,
                            'log_major_color': tmp.xgrid.grid_line_color,
                            'log_minor_visible': tmp.xgrid.minor_grid_line_color is not None,
                            'log_minor_width': tmp.xgrid.minor_grid_line_width,
                            'log_minor_color': tmp.xgrid.grid_line_color,
                            'log_minor_num': tmp.xaxis[0].ticker.num_minor_ticks,
                            'z_major_visible': tmp.ygrid.visible,
                            'z_major_num': tmp.yaxis[0].ticker.desired_num_ticks-1,
                            'z_major_width': tmp.ygrid.grid_line_width,
                            'z_major_color': tmp.ygrid.grid_line_color,
                            'z_minor_visible': tmp.ygrid.minor_grid_line_color is not None,
                            'z_minor_width': tmp.ygrid.minor_grid_line_width,
                            'z_minor_color': tmp.ygrid.grid_line_color,
                            'z_minor_num': tmp.yaxis[0].ticker.num_minor_ticks
                           }
        self.track_props['log_major_dash'] = self.linedash2str(tmp.xgrid.grid_line_dash)
        self.track_props['log_minor_dash'] = self.linedash2str(tmp.xgrid.minor_grid_line_dash)
        self.track_props['z_major_dash'] = self.linedash2str(tmp.ygrid.grid_line_dash)
        self.track_props['z_minor_dash'] = self.linedash2str(tmp.ygrid.minor_grid_line_dash)

    def _initlogprops(self):
        lognames = self.well.getLogNames()
        for log in lognames:
            self.log_props[log] = {'left': None,
                                   'right': None,
                                   'width': 2,
                                   'color': "#%06x" % random.randint(0,0xFFFFFF),
                                   'dash': 'solid'
                                   }

    def _init_log_ui(self):
        lognames = self.well.getLogNames()
        line_props = LinePropertyWidget()
        self.log_fields= {'log': bm.Select(title='Log Display Styling:', options=lognames),
                          'left': bm.Spinner(title='Left Limit:'),
                          'right': bm.Spinner(title='Right Limit:'),
                         }
        self.log_fields.update(line_props.fields)
        self.log_fields['log'].on_change('value', self._update_log_selection)
        self.log_select.js_link('value',self.log_fields['log'], 'options')
        return bl.layout(self.log_fields['log'],
                         self.log_fields['left'],
                         self.log_fields['right'],
                         line_props.layout,
                         align=('center','start')
                        )

    def _initui(self):
        lognames = self.well.getLogNames()
        self.log_select = bm.MultiSelect(title="Logs:", options=lognames, align='center',
                                         height_policy='max')
        self.marker_select = bm.MultiSelect(title="Markers:", options=self.well.getMarkers().data['name'],
                                            align='center', height_policy='max')

        self.track_fields['plot_width'] = bm.Spinner(title='Plot Width(px):', value=self.track_props['plot_width'])
        self.track_fields['background_fill_color'] = bm.ColorPicker(title='Track Background Color:', color=self.track_props['background_fill_color'])
        major_log_grid = GridPropertyWidget('Major Log Grid')
        self.track_fields['log_major'] = major_log_grid.fields
        minor_log_grid = GridPropertyWidget('Minor Log Grid')
        self.track_fields['log_minor'] = minor_log_grid.fields
        major_z_grid = GridPropertyWidget('Major Z Grid')
        self.track_fields['z_major'] = major_z_grid.fields
        minor_z_grid = GridPropertyWidget('Minor Z Grid')
        self.track_fields['z_minor'] = minor_z_grid.fields
        self._update_track_ui()

        self.data_ui = bl.column(self.log_select,
                                 self.marker_select,
                                 self._init_log_ui(),
                                 align=('center','start')
                                )
        self.settings_ui = bl.column(self.track_fields['plot_width'],
                                     self.track_fields['background_fill_color'],
                                     major_log_grid.layout,
                                     minor_log_grid.layout,
                                     major_z_grid.layout,
                                     minor_z_grid.layout,
                                     align=('center','start')
                                    )

    def _update_track_ui(self):
        self.track_fields['plot_width'].update(value=self.track_props['plot_width'])
        self.track_fields['background_fill_color'].update(color=self.track_props['background_fill_color'])

        major_log_fields = self.track_fields['log_major']
        if self.track_props['log_major_visible']:
            major_log_fields['visible'].update(active=[0])
        else:
            major_log_fields['visible'].update(active=[])
        major_log_fields['steps'].update(value=self.track_props['log_major_num'])
        major_log_fields['width'].update(value=self.track_props['log_major_width'])
        major_log_fields['color'].update(color=self.track_props['log_major_color'])
        major_log_fields['dash'].update(value=self.track_props['log_major_dash'])

        minor_log_fields = self.track_fields['log_minor']
        if self.track_props['log_minor_visible']:
            minor_log_fields['visible'].update(active=[0])
        else:
            minor_log_fields['visible'].update(active=[])
        minor_log_fields['steps'].update(value=self.track_props['log_minor_num'])
        minor_log_fields['width'].update(value=self.track_props['log_minor_width'])
        minor_log_fields['color'].update(color=self.track_props['log_minor_color'])
        minor_log_fields['dash'].update(value=self.track_props['log_minor_dash'])

        major_z_fields = self.track_fields['z_major']
        if self.track_props['z_major_visible']:
            major_z_fields['visible'].update(active=[0])
        else:
            major_z_fields['visible'].update(active=[])
        major_z_fields['steps'].update(value=self.track_props['z_major_num'])
        major_z_fields['width'].update(value=self.track_props['z_major_width'])
        major_z_fields['color'].update(color=self.track_props['z_major_color'])
        major_z_fields['dash'].update(value=self.track_props['z_major_dash'])

        minor_z_fields = self.track_fields['z_minor']
        if self.track_props['z_minor_visible']:
            minor_z_fields['visible'].update(active=[0])
        else:
            minor_z_fields['visible'].update(active=[])
        minor_z_fields['steps'].update(value=self.track_props['z_minor_num'])
        minor_z_fields['width'].update(value=self.track_props['z_minor_width'])
        minor_z_fields['color'].update(color=self.track_props['z_minor_color'])
        minor_z_fields['dash'].update(value=self.track_props['z_minor_dash'])

    def linedash2str(self, linedash):
        dash = 'solid'
        for (key, value) in DashPattern._dash_patterns.items():
         if value==linedash:
            dash = key
            break
        return dash

    def _update_log_selection(self, attr, old, new):
        lognm = new
        left = self.log_props[lognm]['left']
        right = self.log_props[lognm]['right']
        if left is None or right is None:
            self.addLog(lognm)
        self._update_log_ui()

    def _update_log_ui(self):
        lognm = self.log_fields['log'].value
        left = self.log_props[lognm]['left']
        right = self.log_props[lognm]['right']
        nicestep = niceNumber(abs(right-left)/10)
        self.log_fields['right'].update(step=nicestep)
        self.log_fields['left'].update(step=nicestep)
        for (key, prop) in self.log_props[lognm].items():
            if key == 'color':
                self.log_fields[key].update(color=prop)
            else:
                self.log_fields[key].update(value=prop)

    def _update_log_props(self):
        lognm = self.log_fields['log'].value
        for (key, field) in self.log_fields.items():
            if key == 'color':
                self.log_props[lognm][key] = field.color
            else:
                self.log_props[lognm][key] = field.value

        start = self.log_fields['left'].value
        end = self.log_fields['right'].value
        color = self.log_fields['color'].color
        width = int(self.log_fields['width'].value)
        dash = self.log_fields['dash'].value
        self.track.extra_x_ranges[lognm].update(start=start, end=end)
        line = self.track.select(name=lognm).glyph
        line.update(line_width=width, line_color=color, line_dash=dash)
        axis = self.track.select(axis_label=lognm)
        axis.update(axis_line_color=color, major_tick_line_color=color,
                    axis_line_width=width,major_tick_line_width=width,
                    axis_line_dash=dash, ticker=(start, end))

    def _update_track_props(self):
        self.track_props['plot_width'] = self.track_fields['plot_width'].value
        self.track_props['background_fill_color'] = self.track_fields['background_fill_color'].color
        self.track_props['log_major_visible'] = self.track_fields['log_major']['visible'].active==[0]
        self.track_props['log_major_num'] = self.track_fields['log_major']['steps'].value
        self.track_props['log_major_width'] = self.track_fields['log_major']['width'].value
        self.track_props['log_major_color'] = self.track_fields['log_major']['color'].color
        self.track_props['log_major_dash'] = self.track_fields['log_major']['dash'].value
        self.track_props['log_minor_visible'] = self.track_fields['log_minor']['visible'].active==[0]
        self.track_props['log_minor_num'] = self.track_fields['log_minor']['steps'].value
        self.track_props['log_minor_width'] = self.track_fields['log_minor']['width'].value
        self.track_props['log_minor_color'] = self.track_fields['log_minor']['color'].color
        self.track_props['log_minor_dash'] = self.track_fields['log_minor']['dash'].value
        self.track_props['z_major_visible'] = self.track_fields['z_major']['visible'].active==[0]
        self.track_props['z_major_num'] = self.track_fields['z_major']['steps'].value
        self.track_props['z_major_width'] = self.track_fields['z_major']['width'].value
        self.track_props['z_major_color'] = self.track_fields['z_major']['color'].color
        self.track_props['z_major_dash'] = self.track_fields['z_major']['dash'].value
        self.track_props['z_minor_visible'] = self.track_fields['z_minor']['visible'].active==[0]
        self.track_props['z_minor_num'] = self.track_fields['z_minor']['steps'].value
        self.track_props['z_minor_width'] = self.track_fields['z_minor']['width'].value
        self.track_props['z_minor_color'] = self.track_fields['z_minor']['color'].color
        self.track_props['z_minor_dash'] = self.track_fields['z_minor']['dash'].value

    def apply_track_props(self):
        self.track.update(plot_width=self.track_props['plot_width'])
        self.track.update(background_fill_color=self.track_props['background_fill_color'])
        self.track.xgrid.visible = self.track_props['log_major_visible']
        if self.track_props['log_major_visible']:
            self.track.xgrid.grid_line_color = self.track_props['log_major_color']
        else:
            self.track.xgrid.grid_line_color = None
        self.track.xaxis[0].ticker.desired_num_ticks=self.track_props['log_major_num']+1
        self.track.xgrid.grid_line_width = self.track_props['log_major_width']
        self.track.xgrid.grid_line_dash = self.track_props['log_major_dash']
        if self.track_props['log_minor_visible']:
            self.track.xgrid.minor_grid_line_color = self.track.xgrid.grid_line_color
        else:
            self.track.xgrid.minor_grid_line_color = None
        self.track.xaxis[0].ticker.num_minor_ticks=self.track_props['log_minor_num']
        self.track.xgrid.minor_grid_line_width = self.track_props['log_minor_width']
        self.track.xgrid.minor_grid_line_dash = self.track_props['log_minor_dash']

        self.track.ygrid.visible = self.track_props['z_major_visible']
        if self.track_props['z_major_visible']:
            self.track.ygrid.grid_line_color = self.track_props['z_major_color']
        else:
            self.track.ygrid.grid_line_color = None
        self.track.yaxis[0].ticker.desired_num_ticks=self.track_props['z_major_num']+1
        self.track.ygrid.grid_line_width = self.track_props['z_major_width']
        self.track.ygrid.grid_line_dash = self.track_props['z_major_dash']
        if self.track_props['z_minor_visible']:
            self.track.ygrid.minor_grid_line_color = self.track.ygrid.grid_line_color
        else:
            self.track.ygrid.minor_grid_line_color = None
        self.track.yaxis[0].ticker.num_minor_ticks=self.track_props['z_minor_num']
        self.track.ygrid.minor_grid_line_width = self.track_props['z_minor_width']
        self.track.ygrid.minor_grid_line_dash = self.track_props['z_minor_dash']

    def _change_tab(self, attr, old, new):
        if old==0:
            self._update_log_props()
        elif old==2:
            self._update_track_props()

        if new==1:
            currange = self.track.y_range
            self._inittrack()
            self.track.update(y_range=currange)
            self.addLogs(self.log_select.value)
            self.addMarkers(self.marker_select.value)
            if self.tracklayout:
                self.tracklayout.children[0] = self.track
        elif new==0:
            self._update_log_ui()
        elif new==2:
            self._update_track_ui()


    def display(self):
        if self.withui:
          self.tracklayout = bl.column(self.track)
          tracktab = bm.Panel(child=self.tracklayout, title="Plot")
          datatab = bm.Panel(child=self.data_ui, title="Data")
          settingstab = bm.Panel(child=self.settings_ui, title="Settings")
          self.tabs = bm.Tabs(tabs=[datatab, tracktab, settingstab])
          self.tabs.on_change('active', self._change_tab)
          return self.tabs
        else:
          return self.track

    def _getLogLimits(self, lognm):
        log = self.well.getLog(lognm)
        left = self.log_props[lognm]['left']
        right = self.log_props[lognm]['right']
        limits = (left, right)
        if left is None or right is None:
            limits = niceRange(min(log.data[lognm]), max(log.data[lognm]))
            self.log_props[lognm]['left'] = limits[0]
            self.log_props[lognm]['right'] = limits[-1]
        return limits

    def addLog(self, lognm):
        log = self.well.getLog(lognm)
        limits = self._getLogLimits(lognm)
        width = self.log_props[lognm]['width']
        color = self.log_props[lognm]['color']
        dash = self.log_props[lognm]['dash']

        axis_style = {'axis_line_color': color,
                      'axis_line_width': width,
                      'axis_line_dash': dash,
                      'major_tick_line_color': color,
                      'major_tick_line_width': width,
                      'major_tick_in': 0,
                      'axis_label_standoff': -10
                     }
        log_style = {'line_color': color,
                     'line_width': width,
                     'line_dash': dash
                    }
        self.track.extra_x_ranges[lognm] = bm.Range1d(limits[0], limits[-1])
        self.track.add_layout(bm.LinearAxis(x_range_name=lognm,axis_label=lognm,ticker=limits,**axis_style), 'above')
        self.track.line(x=lognm, y='depth',x_range_name=lognm,source=log, name=lognm,  **log_style)

    def addLogs(self, logs):
        lognms = self.well.getLogNames()
        showlogs = list(set.intersection(set(logs), set(lognms)))
        for lognm in showlogs:
            self.addLog(lognm)
        if self.withui:
            self.log_select.update(value=showlogs)
            self.log_fields['log'].update(options=showlogs, value=showlogs[0])
            self._update_log_ui()

    def addMarker(self, name, depth, color):
        marker_style = {'line_color': color,
                        'line_width': 2,
                        'line_dash': 'solid'
                       }
        xr = self.track.x_range
        self.track.add_layout(bm.Arrow(end=None,x_start=xr.start,y_start=depth,x_end=xr.end,y_end=depth,**marker_style))
        self.track.add_layout(bm.Label(x=xr.start,y=depth,text=name,text_color=color,render_mode='canvas'))

    def addMarkers(self, markers=None):
        if not markers or len(markers)==0:
            return
        mrksource = self.well.getMarkers().data
        showmarkers = []
        for (name,depth,color) in zip( mrksource['name'],mrksource['depth'],mrksource['color']):
            if name in markers:
                self.addMarker(name,depth,color)
                showmarkers += [name]
        if self.withui:
          self.marker_select.update(value=showmarkers)

    def clearShading(self):
        self.shading = self.back_color

    def addLeftShade(self,log,color):
        pass



#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# AUTHOR   : A. Huck
# DATE     : Nov 2018
#
# tools for reading hdf5 files for NN training
#

from os import path
import random
import json
import numpy as np

import odpy.hdf5 as odhdf5
from odpy.common import std_msg

from dgbpy import dgbscikit
from dgbpy.keystr import *

hdf5ext = 'h5'

def dictAddIfNew( newset, toadd ):
  ret = toadd
  for itmnm in newset:
    toadd.update({itmnm: newset[itmnm]})
  return ret

def getCubeLetNames( info ):
  examples = info[exampledictstr]
  ret = {}
  for groupnm in examples:
    example = examples[groupnm]
    ret.update({groupnm: getCubeLetNamesByGroup(info,groupnm,example)} )
  return ret

def getCubeLetNamesByGroup( info, groupnm, example ):
  collection = example[collectdictstr]
  ret = {}
  for collnm in collection:
    itmidx = collection[collnm][iddictstr]
    ret.update({collnm: getCubeLetNamesByGroupByItem(info,groupnm,collnm,itmidx)})
  return ret

def getCubeLetNamesByGroupByItem( info, groupnm, collnm, idx ):
  h5file = odhdf5.openFile( info[filedictstr], 'r' )
  if not groupnm in h5file:
    return {}
  group = h5file[groupnm]
  dsetnms = list(group.keys())
  if collnm in dsetnms:
    ret = np.arange(len(group[collnm][xdatadictstr]))
  else:
    dsetwithinp = np.chararray.startswith( dsetnms, str(idx)+':' )
    ret = np.extract( dsetwithinp, dsetnms )
  h5file.close()
  return np.ndarray.tolist(ret)

def getGroupSize( filenm, groupnm ):
  h5file = odhdf5.openFile( filenm, 'r' )
  group = h5file[groupnm]
  size = len(group)
  h5file.close()
  return size

def getNrAttribs( info ):
  input = info[inputdictstr]
  collection = input[next(iter(input))][collectdictstr]
  return len(collection)

def getNrOutputs( info ):
  return len( getMainOutputs(info) )

def get_np_shape( shape, nrpts=None, nrattribs=None ):
  ret = ()
  if nrpts != None:
    ret += (nrpts,)
  if nrattribs != None:
    ret += (nrattribs,)
  if isinstance( shape, int ):
    ret += ( 1,1,shape, )
    return ret
  for i in shape:
    ret += (i,)
  return ret

def isRegression( info ):
  return not isClassification( info )

def isClassification( info ):
  if instance(info, dict):
    return info[classdictstr]
  return info == classdatavalstr

def isSeisClass( info ):
  if isinstance(info,dict):
    return info[learntypedictstr] == seisclasstypestr
  return info == seisclasstypestr

def isLogInput( info ):
  if isinstance(info,dict):
    return info[learntypedictstr] == loglogtypestr
  return info == loglogtypestr

def isLogOutput( info ):
  if isinstance(info,dict):
    return info[learntypedictstr] == loglogtypestr or info[learntypedictstr] == seisproptypestr
  return info == loglogtypestr or info == seisproptypestr

def isImg2Img( info ):
  if isinstance(info,dict):
    return info[learntypedictstr] == seisimgtoimgtypestr
  return info == seisimgtoimgtypestr

def isModel( info ):
  return plfdictstr in info

def getCubeLets( infos, collection, groupnm ):
  if len(collection)< 1:
    return {}
  inpnrattribs = getNrAttribs( infos )
  inpshape = infos[inpshapedictstr]
  outshape = infos[outshapedictstr]
  nroutputs = getNrOutputs( infos )
  isclass = infos[classdictstr]
  img2img = isImg2Img( infos )
  examples = infos[exampledictstr]
  if img2img:
    outnrattribs = 1
  outdtype = np.float32
  if isclass:
    outdtype = np.uint8
  h5file = odhdf5.openFile( infos[filedictstr], 'r' )
  group = h5file[groupnm]
  dsetnms = list(group.keys())

  firstcollnm = next(iter(collection))
  hasdata = None
  if firstcollnm in group:
    allcubelets = list()
    alloutputs = list()
    for collnm in collection:
      x_data = group[collnm][xdatadictstr]
      y_data = group[collnm][ydatadictstr]
      dsetnms = collection[collnm]
      nrpts = len(dsetnms)
      inparrshape = get_np_shape(inpshape,nrpts,inpnrattribs)
      if img2img:
        outarrshape = get_np_shape(outshape,nrpts,outnrattribs)
      if len(x_data) == nrpts and len(y_data) == nrpts:
        cubelets = np.resize( x_data, inparrshape ).astype( np.float32 )
        if img2img:
          output = np.resize( y_data, outarrshape ).astype( outdtype )
        else:
          output = np.resize( y_data, (nrpts,nroutputs) ).astype( outdtype )
      else:
        cubelets = np.empty( inparrshape, np.float32 )
        if img2img:
          output = np.empty( outarrshape, outdtype )
        else:
          output = np.empty( (nrpts,nroutputs), outdtype )
        for idx,dsetnm in zip(range(len(dsetnms)),dsetnms):
          dset = x_data[dsetnm]
          odset = y_data[dsetnm]
          cubelets[idx] = np.resize( dset, cubelets[idx].shape )
          if img2img:
            output[idx] = np.resize( odset, output[idx].shape )
          else:
            output[idx] = np.asarray( odset )
      if nrpts > 0:
        allcubelets.append( cubelets )
        alloutputs.append( output )
    if len(allcubelets) > 0:
      cubelets = np.concatenate( allcubelets )
      hasdata = True
    if len(alloutputs) > 0:
      output = np.concatenate( alloutputs )
  else:
    allcubelets = list()
    alloutputs = list()
    for collnm in collection:
      dsetnms = collection[collnm]
      nrpts = len(dsetnms)
      inparrshape = get_np_shape(inpshape,nrpts,inpnrattribs)
      cubelets = np.empty( inparrshape, np.float32 )
      if img2img:
        outarrshape = get_np_shape(outshape,nrpts,outnrattribs)
        output = np.empty( outarrshape, outdtype )
      else:
        output = np.empty( (nrpts,nroutputs), outdtype )
      for idx,dsetnm in zip(range(len(dsetnms)),dsetnms):
        dset = group[dsetnm]
        if img2img:
          try:
            cubelets[idx] = np.resize(dset[:-1],cubelets[idx].shape)
            output[idx] = np.resize(dset[-1],output[idx].shape)
          except Exception as e:
            cubelets[idx] = np.zeros( cubelets[idx].shape, cubelets.dtype )
            output[idx] = np.zeros( output[idx].shape, output[idx].dtype )
        else:
          cubelets[idx] = np.resize(dset,cubelets[idx].shape)
          if isclass :
            output[idx] = odhdf5.getIArray( dset, valuestr )
          else:
            output[idx] = odhdf5.getDArray( dset, valuestr )
      if nrpts > 0:
        allcubelets.append( cubelets )
        alloutputs.append( output )
    if len(allcubelets) > 0:
      cubelets = np.concatenate( allcubelets )
      hasdata = True
    if len(alloutputs) > 0:
      output = np.concatenate( alloutputs )
  h5file.close()
  if not hasdata:
    return {}
  return {
    xtraindictstr: cubelets,
    ytraindictstr: output
  }

def getDatasets_( infos, datasets, fortrain ):
  dictkeys = list()
  if fortrain:
    dictkeys.append( xtraindictstr )
    dictkeys.append( ytraindictstr )
  else:
    dictkeys.append( xvaliddictstr )
    dictkeys.append( yvaliddictstr )
  ret = {}
  cubelets = list()
  for groupnm in datasets:
    cubes = getCubeLets(infos,datasets[groupnm],groupnm)
    if len(cubes) > 0:
      cubelets.append( cubes )
  allx = list()
  ally = list()
  for cubelet in cubelets:
    allx.append( cubelet[xtraindictstr] )
    ally.append( cubelet[ytraindictstr] )
  if len(allx) > 0:
    ret.update({dictkeys[0]: np.concatenate( allx )})
    ret.update({dictkeys[1]: np.concatenate( ally )})
  return ret

def getDatasets( infos, dsetsel=None, train=True, validation=True ):
  ret = {}
  if dsetsel == None:
    datasets = infos[datasetdictstr]
  else:
    datasets = dsetsel
  if train:
    if traindictstr in datasets:
      traindsets = datasets[traindictstr]
    else:
      traindsets = datasets
    trainret = getDatasets_( infos, traindsets, True )
    if len(trainret) > 0:
      for ex in trainret:
        ret.update({ex: trainret[ex]})
  if validation and validdictstr in datasets:
    validret = getDatasets_( infos, datasets[validdictstr], False )
    if len(validret) > 0:
      for ex in validret:
        ret.update({ex: validret[ex]})
  return ret

def validInfo( info ):
  try:
    learntype = odhdf5.getText(info,typestr)
  except KeyError:
    std_msg("No type found. Probably wrong type of hdf5 file")
    return False
  return True

def getInfo( filenm, quick ):
  h5file = odhdf5.openFile( filenm, 'r' )
  info = odhdf5.getInfoDataSet( h5file )
  if not validInfo( info ):
    h5file.close()
    return {}

  learntype = odhdf5.getText(info,typestr)
  isclassification = isSeisClass( learntype )
  if odhdf5.hasAttr(info,contentvalstr):
    isclassification = odhdf5.getText(info,contentvalstr) == classdatavalstr
  img2img = isImg2Img(learntype)
  logoutp = isLogOutput(learntype)

  extxt = 'Examples.'
  ex_sz = odhdf5.getIntValue(info, extxt+'Size') 
  idx = 0
  examples = {}
  while idx < ex_sz:
    exidxstr  = extxt+str(idx)+'.'
    example_sz = odhdf5.getIntValue( info, exidxstr+'Size' )
    exxobj = {
      targetdictstr: odhdf5.getText( info, exidxstr+'Target' ),
      iddictstr: idx
    }
    exsurv = odhdf5.getText(info, exidxstr+'Survey' )
    if len(exsurv) > 0:
      exxobj.update({ locationdictstr: exsurv})
    excompnrstr = exidxstr+'Component'
    if odhdf5.hasAttr(info,excompnrstr):
      exxobj.update({componentdictstr: odhdf5.getIntValue(info,excompnrstr)})
    collection = {}
    idy = 0
    while idy < example_sz:
      exidystr = exidxstr+str(idy)+'.'
      collnm = odhdf5.getText(info,exidystr+'Name')
      collnm = collnm.replace( '/', '_' )
      exxyobj = {
        dbkeydictstr: odhdf5.getText(info, exidystr+'ID' ),
        iddictstr: idy, 
      }
      classnmstr = exidystr+'Class Name'
      if odhdf5.hasAttr(info,classnmstr):
        exxyobj.update({classnmdictstr: odhdf5.getText(info,classnmstr)})
      gidstr = exidystr+'GeomID'
      if odhdf5.hasAttr(info,gidstr):
        exxyobj.update({geomiddictstr: odhdf5.getIntValue(info,gidstr)})
      linenmstr = exidystr+'Line name'
      if odhdf5.hasAttr(info,linenmstr):
        collnm = odhdf5.getText(info,linenmstr)
        if len(collnm) < 1:
          idy = idy+1
          continue
      collection.update({ collnm: exxyobj })
      idy += 1
    if len(collection) > 0:
      exxobj.update({collectdictstr: collection})
    examples.update({ odhdf5.getText(info,exidxstr+'Name'): exxobj })
    idx += 1

  inptxt = 'Input.'
  inp_sz = odhdf5.getIntValue(info,inptxt+'Size')
  idx = 0
  inputs = {}
  while idx < inp_sz:
    inpidxstr = inptxt+str(idx)+'.'
    input_sz = odhdf5.getIntValue( info, inpidxstr+'Size' )
    collection = {}
    idy = 0
    scales = list()
    means = list()
    while idy < input_sz:
      inpidystr = inpidxstr+str(idy)+'.'
      collnm = odhdf5.getText(info, inpidystr+'Name' )
      inpxyobj = {
        iddictstr: idy
      }
      compkey = inpidystr+'Component'
      if odhdf5.hasAttr(info,compkey):
        collnm += '`' + odhdf5.getText(info,compkey)
      idkey = inpidystr+'ID'
      if odhdf5.hasAttr(info,idkey):
        inpxyobj.update({dbkeydictstr: odhdf5.getText(info,idkey)})
      scalekey = inpidystr+'Stats'
      if odhdf5.hasAttr(info,scalekey):
        scaletxt = odhdf5.getAttr(info,scalekey)
        if scaletxt != '0`0':
          scale = odhdf5.getDInterval(info,scalekey)
          means.append( scale[0] )
          scales.append( scale[1] )
      collection.update({ collnm: inpxyobj })
      idy += 1
    inpobj = {
      collectdictstr: collection,
      iddictstr: idx
    }
    inpsurv = odhdf5.getText(info, inpidxstr+'Survey' )
    if len(inpsurv) > 0:
      inpobj.update({ locationdictstr: inpsurv})
    if len(scales) > 0:
      inpobj.update({scaledictstr: dgbscikit.getNewScaler(means,scales) })
    inputs.update({ odhdf5.getText( info, inpidxstr+'Name' ): inpobj})
    idx += 1

  if not odhdf5.hasAttr(info,inpshapestr) or \
     not odhdf5.hasAttr(info,outshapestr):
     raise KeyError

  inpshape = odhdf5.getIArray( info, inpshapestr )
  outshape = odhdf5.getIArray( info, outshapestr )

  retinfo = {
    learntypedictstr: learntype,
    inpshapedictstr: inpshape,
    outshapedictstr: outshape,
    classdictstr: isclassification,
    interpoldictstr: odhdf5.getBoolValue(info,"Edge extrapolation"),
    exampledictstr: examples,
    inputdictstr: inputs,
    filedictstr: filenm
  }

  if not quick:
    retinfo.update({
      datasetdictstr: getCubeLetNames( retinfo )
      })
    
  if odhdf5.hasAttr(info,'Model.Type' ):
    retinfo.update({plfdictstr: odhdf5.getText(info,'Model.Type')})
  if  odhdf5.hasAttr(info,versionstr):
    retinfo.update({versiondictstr: odhdf5.getText(info,versionstr)})
  h5file.close()

  if isLogOutput( learntype ):
    return getWellInfo( retinfo, filenm )
  elif isSeisClass(learntype) or isImg2Img(learntype):
    return getAttribInfo( retinfo, filenm )

  std_msg( "Unrecognized learn type: ", learntype )
  raise KeyError

def getAttribInfo( info, filenm ):
  if info[classdictstr]:
    if isSeisClass(info):
      (classidxs,classnms) = getClassIndices(info)
      classidxs = np.array(classidxs)+1
      info.update( {
        classesdictstr: classidxs.tolist(),
        classnmdictstr: classnms
      })
    else:
      info.update( {classesdictstr: getClassIndicesFromData(info)} )
      
  if not isModel(info):
    info.update( {estimatedsizedictstr: getTotalSize(info)} )
  return info

def getWellInfo( info, filenm ):
  if info[classdictstr]:
    info.update( {
      classesdictstr: getClassIndicesFromData(info),
      classnmdictstr: getMainOutputs(info)[0]
    })

  if not isModel(info):
    info.update( {estimatedsizedictstr: getTotalSize(info)} )
  h5file = odhdf5.openFile( filenm, 'r' )
  infods = odhdf5.getInfoDataSet( h5file )
  zstep = odhdf5.getDValue(infods,"Z step") 
  marker = (odhdf5.getText(infods,"Top marker"),
            odhdf5.getText(infods,"Bottom marker"))
  h5file.close()
  info.update({
    zstepdictstr: zstep,
    rangedictstr: marker,
  })
  return info

def getNrClasses(info):
  return len(info[classesdictstr])

def arroneitemsize( dtype ):
  arr = np.empty(1,dtype)
  return arr.itemsize

def getTotalSize( info ):
  inpnrattribs = getNrAttribs( info )
  inpshape = info[inpshapedictstr]
  outshape = info[outshapedictstr]
  examples = info[exampledictstr]
  h5file = odhdf5.openFile( info[filedictstr], 'r' )
  nrpts = 0
  for groupnm in examples:
    grp = h5file[groupnm]
    collection = examples[groupnm][collectdictstr]
    firstcollnm = next(iter(collection))
    if not firstcollnm in grp:
      nrpts += len(grp)
      continue
    for collnm in collection:
      nrpts += len(grp[collnm][xdatadictstr]) 
  h5file.close()
  examplesshape = get_np_shape( inpshape, nrpts, inpnrattribs )
  x_size = np.prod( examplesshape, dtype=np.int64 ) * arroneitemsize( np.float32 )
  if info[classdictstr]:
    nroutvals = getNrClasses( info )
  else:
    nroutvals = getNrOutputs( info )
  outshape = get_np_shape( outshape, nrpts, nroutvals )
  y_size = np.prod( outshape, dtype=np.int64 ) * arroneitemsize( np.float32 )
  return x_size + y_size

modeloutstr = 'Model.Output.'
def modelIdxStr( idx ):
  return modeloutstr + str(idx) + '.Name'

def addInfo( inpfile, plfnm, filenm, infos, clssnm ):
  h5filein = odhdf5.openFile( inpfile, 'r' )
  h5fileout = odhdf5.openFile( filenm, 'r+' )
  dsinfoin = odhdf5.getInfoDataSet( h5filein )
  dsinfoout = odhdf5.ensureHasDataset( h5fileout )
  attribman = dsinfoin.attrs
  for attribkey in attribman:
    dsinfoout.attrs[attribkey] = attribman[attribkey]
  h5filein.close()
  odhdf5.setAttr( dsinfoout, versionstr, str(1) )
  odhdf5.setAttr( dsinfoout, 'Model.Type', plfnm )
  odhdf5.setAttr( dsinfoout, 'Model.Class', clssnm )
  if plfnm == kerasplfnm:
    #The model size may be smaller than from the example data
    odhdf5.setArray( dsinfoout, inpshapestr, infos[inpshapedictstr] )
    if isImg2Img( infos ):
      odhdf5.setArray( dsinfoout, outshapestr, infos[outshapedictstr] )
    else:
      odhdf5.setArray( dsinfoout, outshapestr, getNrOutputs(infos) )

  outps = getOutputs( infos )
  odhdf5.setAttr( dsinfoout, modeloutstr+'Size', str(len(outps)) )
  for idx,outp in zip(range(len(outps)),outps):
    odhdf5.setAttr( dsinfoout, modelIdxStr(idx), outp )

  inp = infos[inputdictstr]
  for inputnm in inp:
    input = inp[inputnm]
    if not scaledictstr in input:
      continue
    scale = input[scaledictstr]
    keyval = 'Input.' + str(input[iddictstr]) + '.'
    for i in range(len(scale.scale_)):
      keyvali = keyval+str(i)+'.Stats'
      odhdf5.setArray( dsinfoout, keyvali, [scale.mean_[i], scale.scale_[i]] )

  h5fileout.close()

def getClassIndices( info, filternms=None ):
  if isLogOutput( info ):
    return ([],[]) #No support (yet?)
  allclasses = {}
  examples = info[exampledictstr]
  for groupnm in examples:
    collection = examples[groupnm][collectdictstr]
    collclassnms = {}
    for collnm in collection:
      collitm = collection[collnm]
      if not classnmdictstr in collitm:
        continue
      classnm = collitm[classnmdictstr]
      if filternms==None or classnm in filternms:
        collclassnms.update({classnm: collitm[iddictstr]})
    dictAddIfNew( collclassnms, allclasses )
  idxs = list()
  classnms = list()
  for classnm in allclasses:
    classitm = allclasses[classnm]
    idxs.append( classitm )
    classnms.append( classnm )
  return (idxs,classnms)

def getClassIndicesFromData( info ):
  if classesdictstr in info:
    return info[classesdictstr]
  filenm = info[filedictstr]
  h5file = odhdf5.openFile( filenm, 'r' )
  dsinfoin = odhdf5.ensureHasDataset( h5file )
  if odhdf5.hasAttr( dsinfoin, classesvalstr ):
    return odhdf5.getIArray( dsinfoin, classesvalstr )
  isimg2img = isImg2Img( info )
  groups = info[exampledictstr].keys()
  ret = list()
  for groupnm in groups:
    grp = h5file[groupnm]
    if isimg2img:
      for inpnm in grp:
        sublist = list(set(grp[inpnm][-1].astype('uint8').ravel()))
        sublist.extend( ret )
        ret = list(set(sublist))
    else:
      nrvals = len(grp)
      vals = np.empty( nrvals, dtype='uint8' )
      for ival,dsetnm in zip(range(nrvals),grp):
        dset = grp[dsetnm]
        vals[ival] = odhdf5.getIntValue( dset, valuestr )
      sublist = list(set(vals.ravel()))
      sublist.extend( ret )
      ret = list(set(sublist))
  ret = np.sort( ret )
  h5file.close()
  h5fileout = odhdf5.openFile( filenm, 'r+' )
  dsinfoout = odhdf5.ensureHasDataset( h5fileout )
  odhdf5.setArray( dsinfoout, classesvalstr, ret )
  h5fileout.close()
  return ret

def getMainOutputs( info ):
  examples = info[exampledictstr]
  firstexamplenm = next(iter(examples))
  targets = examples[firstexamplenm][targetdictstr]
  ret = list()
  if isinstance(targets,list):
    ret.extend(targets)
  else:
    ret.append(targets)
  return ret
  
def getOutputs( info ):
  ret = getMainOutputs( info )
  if isSeisClass(info):
    ret.extend( info[classnmdictstr] )
    ret.append( confvalstr )
  return ret

def getOutputNames( filenm, indices ):
  h5file = odhdf5.openFile( filenm, 'r' )
  info = odhdf5.getInfoDataSet( h5file )
  ret = list()
  for idx in indices:
    ret.append( odhdf5.getText(info,modelIdxStr(idx)) )
  h5file.close()
  return ret



import numpy as np
from joblib import load
import json

def model_info( modelfnm ):
    model = load( modelfnm )
    esttype = getattr(model,'_estimator_type','Unknown')
    modtype = getattr(type(model),'__name__', None)
    module = getattr(model,'__module__',None)
    params = json.dumps(model.get_params())
    if hasattr(model,'estimators_'):
      if isinstance(model.estimators_,np.ndarray):
        model = model.estimators_[0][0]
      else:
        model = model.estimators_[0]

    nfeatures = getattr(model,'n_features_',None)
    noutputs = getattr(model,'n_outputs_',None)
    coef = getattr(model,'coef_', None)
    classes = getattr(model,'classes_',None)
    if nfeatures is None:
      if coef is not None:
        nfeatures = coef.shape[1]

    if noutputs is None:
      if coef is not None:
        noutputs = coef.shape[0]

    if classes is not None:
      pass
    elif esttype is 'classifier' and noutputs is not None:
        classes = [i for i in range(noutputs)]

    return (esttype, modtype, module, params, nfeatures, noutputs, classes)



import tensorflow as tf
import keras.backend.tensorflow_backend as tfback

def _get_available_gpus():
  """Get a list of available gpu devices (formatted as strings).
  
    Workaround for TensorFlow 2 error when loading TensorFlow 1 models in Keras

    # Returns
      A list of available GPU devices.
  """
  if tfback._LOCAL_DEVICES is None:
    devices = tf.config.list_logical_devices()
    tfback._LOCAL_DEVICES = [x.name for x in devices]
  return [x for x in tfback._LOCAL_DEVICES if 'device:gpu' in x.lower()]

tfback._get_available_gpus = _get_available_gpus 



#__________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Mar 2019
#
# _________________________________________________________________________
# runs a machine learning training a stand-alone process
#

import sys
import os
import platform
import argparse
import json
import traceback as tb

from odpy import common as odcommon
from odpy.oscommand import getPythonExecNm, printProcessTime
import dgbpy.keystr as dgbkeys
import dgbpy.mlapply as dgbmlapply

parser = argparse.ArgumentParser(
            description='Run machine learning model training')
parser.add_argument( '-v', '--version',
            action='version', version='%(prog)s 1.0' )
parser.add_argument( 'h5file',
            type=argparse.FileType('r'),
            help='HDF5 file containing the training data' )
parser.add_argument( '--dict',
            dest='dict', metavar='JSON_DICT', nargs=1,
            help='Dictionary: {"platform": "keras", "output": "<new model>", "parameters": {trainpars}}' )
datagrp = parser.add_argument_group( 'Data' )
datagrp.add_argument( '--dataroot',
            dest='dtectdata', metavar='DIR', nargs=1,
            help='Survey Data Root' )
datagrp.add_argument( '--survey',
            dest='survey', nargs=1,
            help='Survey name' )
odappl = parser.add_argument_group( 'OpendTect application' )
odappl.add_argument( '--dtectexec',
            metavar='DIR', nargs=1,
            help='Path to OpendTect executables' )
loggrp = parser.add_argument_group( 'Logging' )
loggrp.add_argument( '--proclog',
            dest='logfile', metavar='file', nargs='?',
            type=argparse.FileType('w'), default=sys.stdout,
            help='Progress report output' )
loggrp.add_argument( '--syslog',
            dest='sysout', metavar='stdout', nargs='?',
            type=argparse.FileType('a'), default=sys.stdout,
            help='Standard output' )
args = vars(parser.parse_args())
odcommon.initLogging( args )
odcommon.proclog_logger.setLevel( 'DEBUG' )

if __name__ == '__main__':
  odcommon.log_msg( 'Starting program:', getPythonExecNm(), " ".join(sys.argv) )
  odcommon.log_msg( 'Processing on:', platform.node() )
  odcommon.log_msg( 'Process ID:', os.getpid(), '\n' )
  printProcessTime( 'Machine Learning Training', True, print_fn=odcommon.log_msg )
  odcommon.log_msg( '\n' )
  dict = json.loads( args['dict'][0] )
  traintype = None
  if dgbkeys.learntypedictstr in dict:
    traintype = dgbmlapply.TrainType[ dict[dgbkeys.learntypedictstr] ]
  model = dict.get('model', None)
  logdir = dict.get('logdir', None)
  clearlogs = dict.get('cleanlogdir', False)
  try:
    success = dgbmlapply.doTrain( args['h5file'].name,
                                  platform=dict['platform'],
                                  type=traintype,
                                  params=dict['parameters'],
                                  logdir=logdir,
                                  clearlogs=clearlogs,
                                  modelin=model,
                                  outnm=dict['output'],
                                  args=args )
  except Exception as e:
    exc_type, exc_obj, exc_tb = sys.exc_info()
    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
    stackstr = ''.join(tb.extract_tb(exc_tb,limit=10).format())
    odcommon.log_msg( 'Training error exception:' )
    odcommon.log_msg( repr(e), 'on line', exc_tb.tb_lineno, 'of script', fname )
    odcommon.log_msg( stackstr )
    sys.exit(1)
  if not success:
    sys.exit(1)

  odcommon.log_msg( '\n' )
  printProcessTime( 'Machine Learning Training', False, print_fn=odcommon.log_msg )
  odcommon.log_msg( "Finished batch processing.\n" )
  sys.exit(0)



#__________________________________________________________________________
#
# (C) dGB Beheer B.V.; (LICENSE) http://opendtect.org/OpendTect_license.txt
# Author:        A. Huck
# Date:          Jan 2019
#
# _________________________________________________________________________
# various tools for applying machine learning
#

from enum import Enum
import numpy as np
import os

from odpy.common import log_msg, get_log_file
from odpy.oscommand import printProcessTime
import dgbpy.keystr as dgbkeys
import dgbpy.hdf5 as dgbhdf5
import dgbpy.mlio as dgbmlio

TrainType = Enum( 'TrainType', 'New Resume Transfer', module=__name__ )

def computeScaler_( datasets, infos, scalebyattrib ):
  ret = dgbmlio.getTrainingDataByInfo( infos, datasets )
  allx = list()
  if dgbkeys.xtraindictstr in ret:
    if len(ret[dgbkeys.xtraindictstr]) > 0:
      allx.append( ret[dgbkeys.xtraindictstr] )
  if dgbkeys.xvaliddictstr in ret:
    if len(ret[dgbkeys.xvaliddictstr]) > 0:
      allx.append( ret[dgbkeys.xvaliddictstr] )
  if len(allx) > 0:
    x_data = np.concatenate( allx )
    return getScaler( x_data, byattrib=scalebyattrib )
  return None

def computeChunkedScaler_(datasets,infos,groupnm,scalebyattrib):
  chunknb = len(datasets)
  chunkmean = list()
  chunkstd = list()
  chunklen = list()
  for dataset in datasets:
    datasetchunk = dgbmlio.getDatasetsByGroup( dataset, groupnm )
    chunksize = 0
    for keynm in datasetchunk:
      data = datasetchunk[keynm]
      if groupnm in data:
        dsets = data[groupnm]
        for subgroupnm in dsets:
          chunksize += len(dsets[subgroupnm])
    chunklen.append(chunksize)
    scaleronechunk = computeScaler_( datasetchunk, infos, scalebyattrib )
    if scaleronechunk == None:
      chunkmean.append(0)
      chunkstd.append(0)
    else:
      chunkmean.append(scaleronechunk.mean_)
      chunkstd.append(scaleronechunk.scale_)

  if chunknb < 2:
    return getNewScaler( chunkmean[0], chunkstd[0] )
  if scalebyattrib:
    attrnb = dgbhdf5.getNrAttribs(infos)
  else:
    attrnb = 1
  #Calculate Mean and Var
  totalmean = list()
  totalstd = list()
  for attr in range(attrnb):
    attrmeansum = 0
    attrstdsum = 0
    attrsize = 0
    for ichunk in range(chunknb):
      attrchunksize = chunklen[ichunk]
      if attrchunksize > 0:
        attrmeansum += chunkmean[ichunk][attr] * attrchunksize
        attrstdsum += chunkstd[ichunk][attr] * attrchunksize
      attrsize += attrchunksize
    if attrsize > 0:
      attrmeansum /= attrsize
      attrstdsum /= attrsize
      totalmean.append( attrmeansum )
      totalstd.append( attrstdsum )
  return getNewScaler( totalmean, totalstd )

def computeScaler( infos, scalebyattrib, force=False ):
  datasets = infos[dgbkeys.trainseldicstr]
  inp = infos[dgbkeys.inputdictstr]
  if infos[dgbkeys.learntypedictstr] == dgbkeys.loglogtypestr:
    if not dgbmlio.hasScaler(infos) or force:
      printProcessTime( 'Scaler computation', True, print_fn=log_msg )
      scaler = computeScaler_( datasets[0], infos, scalebyattrib )
      printProcessTime( 'Scaler computation', False, print_fn=log_msg, withprocline=False )
      for groupnm in inp:
        inp[groupnm].update({dgbkeys.scaledictstr: scaler})
  else:
    for groupnm in inp:
      if dgbmlio.hasScaler( infos, groupnm ) and not force:
        continue
      printProcessTime( 'Scaler computation', True, print_fn=log_msg )
      scaler = computeChunkedScaler_(datasets,infos,groupnm,scalebyattrib)
      printProcessTime( 'Scaler computation', False, print_fn=log_msg, withprocline=False )
      inp[groupnm].update({dgbkeys.scaledictstr: scaler})
  return infos

def getScaledTrainingData( filenm, flatten=False, scale=True, force=False, 
                           nbchunks=1, split=None ):
  if isinstance(scale,bool):
    doscale = scale
    scalebyattrib = True
  else:
    doscale = scale[0]
    scalebyattrib = len(scale) < 2 or scale[1]

  infos = dgbmlio.getInfo( filenm )
  dsets = dgbmlio.getChunks(infos[dgbkeys.datasetdictstr],nbchunks)
  datasets = []
  for dset in dsets:
    datasets.append( dgbmlio.getDatasetNms(dset,validation_split=split) )
  infos.update({dgbkeys.trainseldicstr: datasets})
  if doscale:
    infos = computeScaler( infos, scalebyattrib, force )
  if nbchunks > 1: #Decimate, only need to return the updated info
    return {dgbkeys.infodictstr: infos}
  return getScaledTrainingDataByInfo( infos, flatten=flatten, scale=scale )

def getInputList( datasets ):
  ret = {}
  for keynm in datasets:
    dgbhdf5.dictAddIfNew( datasets[keynm], ret )
  return ret.keys()

def getScaledTrainingDataByInfo( infos, flatten=False, scale=True, ichunk=0 ):
  printProcessTime( 'Data pre-loading', True, print_fn=log_msg )
  x_train = list()
  y_train = list()
  x_validate = list()
  y_validate = list()
  datasets = infos[dgbkeys.trainseldicstr][ichunk]
  groups = getInputList( datasets )
  for groupnm in groups:
    dsets = dgbmlio.getDatasetsByGroup( datasets, groupnm )
    ret = dgbmlio.getTrainingDataByInfo( infos, dsets )
    if scale and groupnm in infos[dgbkeys.inputdictstr]:
      scaler = infos[dgbkeys.inputdictstr][groupnm][dgbkeys.scaledictstr]
      if dgbkeys.xtraindictstr in ret:
        transform( ret[dgbkeys.xtraindictstr], scaler )
      if dgbkeys.xvaliddictstr in ret:
        transform( ret[dgbkeys.xvaliddictstr], scaler )
    if dgbkeys.xtraindictstr in ret:
      if len(ret[dgbkeys.xtraindictstr]) > 0:
        x_train.append( ret[dgbkeys.xtraindictstr] )
    if dgbkeys.xvaliddictstr in ret:
      if len(ret[dgbkeys.xvaliddictstr]) > 0:
        x_validate.append( ret[dgbkeys.xvaliddictstr] )
    if dgbkeys.ytraindictstr in ret:
      if len(ret[dgbkeys.ytraindictstr]) > 0:
        y_train.append( ret[dgbkeys.ytraindictstr] )
    if dgbkeys.yvaliddictstr in ret:
      if len(ret[dgbkeys.yvaliddictstr]) > 0:
        y_validate.append( ret[dgbkeys.yvaliddictstr] )
  nrexamples = 0
  if len(x_train)>0:
    x_train = np.concatenate(x_train)
    nrexamples += len(x_train)
    ret.update({dgbkeys.xtraindictstr: x_train })
  if len(y_train)>0:
    ret.update({dgbkeys.ytraindictstr: np.concatenate(y_train) })
  if len(x_validate)>0:
    x_validate = np.concatenate(x_validate)
    nrexamples += len(x_validate)
    ret.update({dgbkeys.xvaliddictstr: x_validate })
  if len(y_validate)>0:
    ret.update({dgbkeys.yvaliddictstr: np.concatenate(y_validate) })

  printProcessTime( 'Data pre-loading', False, print_fn=log_msg, withprocline=False )

  import copy
  decinfos = copy.deepcopy( infos )
  decinfos[dgbkeys.trainseldicstr] = [datasets]
  ret.update({dgbkeys.infodictstr: decinfos})

  if not flatten:
    return ret

  if dgbkeys.xtraindictstr in ret:
    x_train = ret[dgbkeys.xtraindictstr]
    ret[dgbkeys.xtraindictstr] = np.reshape( x_train, (len(x_train),-1) )
  if dgbkeys.xvaliddictstr in ret:
    x_validate = ret[dgbkeys.xvaliddictstr]
    ret[dgbkeys.xvaliddictstr] = np.reshape( x_validate, (len(x_validate),-1) )
  return ret

def getScaler( x_train, byattrib=True ):
  import dgbpy.dgbscikit as dgbscikit
  return dgbscikit.getScaler( x_train, byattrib )

def getNewScaler( mean, scale ):
  import dgbpy.dgbscikit as dgbscikit
  return dgbscikit.getNewScaler( mean, scale )

def transform(x_train,scaler):
  nrattribs = scaler.n_samples_seen_
  if nrattribs > 0:
    for iattr in range(nrattribs):
      if nrattribs == 1:
        inp = x_train
      else:
        inp = x_train[:,iattr]
      inp -= scaler.mean_[iattr]
      doscale = np.flatnonzero( scaler.scale_ )
      if (doscale == iattr)[iattr]:
        inp /= scaler.scale_[iattr]


def doTrain( examplefilenm, platform=dgbkeys.kerasplfnm, type=TrainType.New,
             params=None, outnm=dgbkeys.modelnm, logdir=None, clearlogs=False, modelin=None,
             args=None ):
  (model,infos) = (None,None)
  if type == None:
    type = TrainType.New
  if type != TrainType.New:
    (model,infos) = dgbmlio.getModel( modelin, fortrain=True )

  trainingdp = None
  validation_split = 0.2 #Params?
  if platform == dgbkeys.kerasplfnm:
    import dgbpy.dgbkeras as dgbkeras
    import tempfile
    if params == None:
      params = dgbkeras.getParams()
    dgbkeras.set_compute_device( params[dgbkeras.prefercpustr] )
    
    trainingdp = getScaledTrainingData( examplefilenm, flatten=False,
                                        scale=True, force=False,
                                        nbchunks=params['nbchunk'],
                                        split=validation_split )
    logdir = dgbkeras.getLogDir( examplefilenm, logdir, clearlogs, args )
    if type == TrainType.New:
      model = dgbkeras.getDefaultModel(trainingdp[dgbkeys.infodictstr],
                                       type=params['type'],
                                       learnrate=params['learnrate'])
    elif type == TrainType.Transfer:
      model = dgbkeras.transfer( model )

    tempmodelnm = None
    logfnm = get_log_file()
    if logfnm != None:
      tempmodelfnm = tempfile.NamedTemporaryFile( dir=os.path.dirname(logfnm) )
      tempmodelnm = tempmodelfnm.name + '.h5'
      tempmodelfnm = None
    print('--Training Started--', flush=True)
    try:
      model = dgbkeras.train( model, trainingdp, params=params,
                              trainfile=examplefilenm, logdir=logdir,
                              withaugmentation=dgbkeras.withaugmentation,
                              tempnm=tempmodelnm )
    except ResourceExhaustedError:
      model = dgbmlio.getModel( tempmodelnm, True )
    try:
      if os.path.exists(tempmodelnm):
        os.remove( tempmodelnm )
    except:
      pass
  elif platform == dgbkeys.scikitplfnm:
    import dgbpy.dgbscikit as dgbscikit
    if params == None:
      params = dgbscikit.getParams()
    trainingdp = getScaledTrainingData( examplefilenm, flatten=True,
                                        scale=True, force=False,
                                        split=validation_split )
    if type == TrainType.New:
      model = dgbscikit.getDefaultModel( trainingdp[dgbkeys.infodictstr],
                                         params )
    print('--Training Started--', flush=True)
    model = dgbscikit.train( model, trainingdp )
  else:
    log_msg( 'Unsupported machine learning platform' )
    raise AttributeError

  infos = trainingdp[dgbkeys.infodictstr]
  modtype = dgbmlio.getModelType( infos )
  outfnm = dgbmlio.getSaveLoc( outnm, modtype, args )
  dgbmlio.saveModel( model, examplefilenm, platform, infos, outfnm )
  return (outfnm != None and os.path.isfile( outfnm ))

def reformat( res, applyinfo ):
  if dgbkeys.preddictstr in res:
    res[dgbkeys.preddictstr] = res[dgbkeys.preddictstr].astype( applyinfo[dgbkeys.dtypepred] )
  if dgbkeys.probadictstr in res:
    res[dgbkeys.probadictstr] = res[dgbkeys.probadictstr].astype( applyinfo[dgbkeys.dtypeprob] )
  if dgbkeys.confdictstr in res:
    res[dgbkeys.confdictstr] = res[dgbkeys.confdictstr].astype( applyinfo[dgbkeys.dtypeconf] )
  return res

def doApplyFromFile( modelfnm, samples, outsubsel=None ):
  (model,info) = dgbmlio.getModel( modelfnm, fortrain=False )
  applyinfo = dgbmlio.getApplyInfo( info, outsubsel )
  return doApply( model, info, samples, applyinfo=applyinfo )

def doApply( model, info, samples, scaler=None, applyinfo=None, batchsize=None ):
  platform = info[dgbkeys.plfdictstr]
  if applyinfo == None:
    applyinfo = dgbmlio.getApplyInfo( info )

  isclassification = info[dgbkeys.classdictstr]
  withpred = dgbkeys.dtypepred in applyinfo
  withprobs = False
  doprobabilities = False
  withconfidence = False
  if isclassification:
    if dgbkeys.probadictstr in applyinfo:
      withprobs = applyinfo[dgbkeys.probadictstr]
      doprobabilities = len(withprobs) > 0
    withconfidence = dgbkeys.dtypeconf in applyinfo

  res = None
  if platform == dgbkeys.kerasplfnm:
    import dgbpy.dgbkeras as dgbkeras
    res = dgbkeras.apply( model, samples, isclassification, withpred, withprobs, withconfidence, doprobabilities, \
                          scaler=None, batch_size=batchsize  )
  elif platform == dgbkeys.scikitplfnm:
    import dgbpy.dgbscikit as dgbscikit
    res = dgbscikit.apply( model, samples, scaler, isclassification, withpred, withprobs, withconfidence, doprobabilities )
  elif platform == dgbkeys.numpyvalstr:
    res = numpyApply( samples )
  else:
    log_msg( 'Unsupported machine learning platform' )
    raise AttributeError

  if isclassification and withpred:
    classarr = res[dgbkeys.preddictstr]
    dgbmlio.unnormalize_class_vector( classarr, info[dgbkeys.classesdictstr] )
    res.update({dgbkeys.preddictstr: classarr})

  return reformat( res, applyinfo )

def numpyApply( samples ):
  import numpy as np
  return {
    dgbkeys.preddictstr: np.mean( samples, axis=(1,2,3,4) )
  }

def inputCount( infos, raw=False, dsets=None ):
  if dsets == None:
    if raw or not dgbkeys.trainseldicstr in infos:
      return inputCount_( infos[dgbkeys.datasetdictstr] )
    else:
      dsets = infos[dgbkeys.trainseldicstr]
  if isinstance(dsets,list):
    return inputCountList(infos,dsets)
  ret = {}
  for keynm in dsets:
    ret.update({keynm: inputCount_(dsets[keynm])})
  return ret

def inputCountList( infos, dsetslist ):
  ret = list()
  for dsets in dsetslist:
    ret.append( inputCount(infos,dsets=dsets) )
  return ret

def inputCount_( dsets ):
  ret = {}
  dscounts = dgbmlio.datasetCount( dsets )
  for groupnm in dscounts:
    if groupnm == 'size':
      continue
    ret.update({groupnm: dscounts[groupnm]['size']})
  return ret

def split( arrays, ratio ):
  if len(arrays) < 1:
    return None
  nrpts = len(arrays[0])
  idxs = np.random.shuffle( np.arange(np.int64(nrpts)) )

